#!/bin/csh -f
# job script for L-system of Earth Simulator
# originally written by Masanori Kameyama on 20031020
# modified on 20031021; wrapper script for mpirun was removed
#                       directory names should end in "/"
#             20031204; modified for flat MPI parallelization
#
set SYSTEM = L
@ ProcessorNode = 2
@ ProcessPerNode = 8
set MPIRUN = mpirun
set ELAPSETIME = 10:00
set head = R`date +%Y%m%d%H%M`
set script = ${head}.sh
set DoSubmit = Confirm
set DoF_PROGINF = yes
set DoMPIPROGINF = yes
set DoPROFILER = yes
set DoFTRACE = yes
set QSUB = qsub

# set input/output files
set EXENAME = conv3d.exe
set SHOST = S
set SHOME = /S/home006/y0057
set SDATA = /S/data006/y0057
set MDPSD = /M/disk003/y0057
set EXEDIR = ${SHOME}/conv3d/TEST
set MYSDIR = ${SHOME}/conv3d/TEST
set MYMDIR = ${MDPSD}/conv3d/TEST
set MYLDIR = conv3d/TEST
set LOGDIR = ${SHOME}/conv3d/TEST
set myfiles = `cat finput`
set inparam = $myfiles[1]
set outparam = $myfiles[2]
set logfile = $myfiles[3]
set file9 = $myfiles[4]
set file10 = $myfiles[5]
set file11 = $myfiles[6]
set file20 = $myfiles[7]

# set statistics of file9/10/11
set dum = `cat run_controls`
set iflag9  = $dum[1]
set iflag10 = $dum[2]
set iflag11 = $dum[3]
set file9dir = ${MYSDIR}
set file10dir = ${MYMDIR}
set file11dir = ${MYMDIR}

# set disk/filesystems
set MountHome = yes
set SFSsize = 10gb

# read local configuration
set myname = `basename $0`
set config = $myname:r.conf
if ( -f ${config} ) source ${config}

# set for LastRankID and LastPnID
@ TotalMPINum = ${ProcessorNode} * ${ProcessPerNode}
@ LastRankID = ${TotalMPINum} - 1
@ LastPnID = ${ProcessorNode} - 1

# setting for file9
set tmp4file9 = /tmp/$$_file9
switch ($iflag9)
    case 2:
	cat <<EOF > $tmp4file9
#PBS -I "mdps:${file9dir}/t${file9},${MYLDIR}/"
#PBS -I "mdps:${file9dir}/u${file9},${MYLDIR}/"
#PBS -I "mdps:${file9dir}/v${file9},${MYLDIR}/"
#PBS -I "mdps:${file9dir}/w${file9},${MYLDIR}/"
#PBS -I "mdps:${file9dir}/p${file9},${MYLDIR}/"
#PBS -I "mdps:${file9dir}/q${file9},${MYLDIR}/"
#PBS -I "mdps:${file9dir}/s${file9},${MYLDIR}/"
#PBS -I "mdps:${file9dir}/e${file9},${MYLDIR}/"
#PBS -I "mdps:${file9dir}/h${file9},${MYLDIR}/"
EOF
	breaksw
    case 1:
	cat <<EOF > $tmp4file9
#PBS -I "mdps:${file9dir}/${file9}.p_%04{0-${LastRankID}}n,%{(0-${LastPnID})@${ProcessPerNode}}n:${MYLDIR}/"
EOF
	breaksw
    default:
	cat <<EOF > $tmp4file9
#PBS -I "mdps:${file9dir}/${file9},${MYLDIR}/"
EOF
	breaksw
endsw

# setting for file10
set tmp4file10 = /tmp/$$_file10
switch ($iflag10)
    case 2:
	cat <<EOF >> $tmp4file10
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/t${file10}"
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/u${file10}"
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/v${file10}"
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/w${file10}"
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/p${file10}"
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/q${file10}"
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/s${file10}"
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/e${file10}"
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/h${file10}"
EOF
	breaksw
    case 1:
	cat <<EOF >> $tmp4file10
#PBS -O "mdps:${file10dir}/,%{(0-${LastPnID})@${ProcessPerNode}}n:${MYLDIR}/${file10}.p_%04{0-${LastRankID}}n"
EOF
	breaksw
    default:
	cat <<EOF >> $tmp4file10
#PBS -O "mdps:${file10dir}/,0:${MYLDIR}/${file10}"
EOF
	breaksw
endsw

# setting for file11
set tmp4file11 = /tmp/$$_file11
switch ($iflag11)
    case 1:
	cat <<EOF > $tmp4file11
#PBS -O "mdps:${file11dir}/,%{(0-${LastPnID})@${ProcessPerNode}}n:${MYLDIR}/${file11}.p_%04{0-${LastRankID}}n"
EOF
	breaksw
    default:
	cat <<EOF > $tmp4file11
#PBS -O "mdps:${file11dir}/,0:${MYLDIR}/${file11}"
EOF
    breaksw
endsw

cat <<EOF > $script
#!/bin/sh
#PBS -T mpisx
#PBS -q ${SYSTEM} -b ${ProcessorNode} -l elapstim_req=${ELAPSETIME} -l filecap_job=${SFSsize}
#
# standard output and standard error
#PBS -o mdps:${LOGDIR}/stdout%s.%03j
#PBS -e mdps:${LOGDIR}/stderr%s.%03j
#
EOF

if ( $MountHome =~ [Yy]* ) then
cat <<EOF >> $script
#PBS -v NFS_MNT_ONE=${SHOME}
#
EOF
endif

if ( $DoF_PROGINF =~ [Yy]* ) then
cat <<EOF >> $script
# for proginf; output in stderr
#PBS -w F_PROGINF=YES
#PBS -v F_PROGINF=YES
#
EOF
endif

if ( $DoMPIPROGINF =~ [Yy]* ) then
cat <<EOF >> $script
# for MPIproginf; output in stderr
#PBS -w MPIPROGINF=YES
#PBS -v MPIPROGINF=YES
#
EOF
endif

if ( $DoPROFILER =~ [Yy]* ) then
cat <<EOF >> $script
# for profiler
#PBS -O "mdps:${LOGDIR}/,0:${MYLDIR}/mon.out"
#
EOF
endif

if ($DoFTRACE =~ [Yy]* ) then
cat <<EOF >> $script
# for ftrace
#PBS -O "mdps:${LOGDIR}/,0:${MYLDIR}/ftrace.sdf"
#PBS -O "mdps:${LOGDIR}/,%{(0-${LastPnID})@8}n:${MYLDIR}/ftrace.out.0.%{0-${LastRankID}}n"
EOF
endif

cat <<EOF >> $script
#
#PBS -I "mdps:${EXEDIR}/${EXENAME},${MYLDIR}/"
#PBS -I "mdps:${MYSDIR}/finput,${MYLDIR}/"
#PBS -I "mdps:${MYSDIR}/run_controls,${MYLDIR}/"
#PBS -I "mdps:${MYSDIR}/${inparam},${MYLDIR}/"
#PBS -O "mdps:${MYSDIR}/,0:${MYLDIR}/${outparam}"
#PBS -O "mdps:${MYSDIR}/,0:${MYLDIR}/${file20}"
EOF
cat $tmp4file9 $tmp4file10 $tmp4file11 >> $script
rm -f $tmp4file9 $tmp4file10 $tmp4file11
cat <<EOF >> $script
cd ${MYLDIR}
${MPIRUN} -nnp ${ProcessPerNode} ${EXENAME}
ls -l
EOF

if ( $DoSubmit =~ [Yy]* ) then
    ${QSUB} $script
    exit 0
else if ( $DoSubmit =~ [Cc]* ) then
    ${PAGER} $script
    echo -n "Submit? "
    set ans = $<
    if ($ans =~ [Yy]*) ${QSUB} $script
    exit 0
else
    ${PAGER} $script
    rm -f $script
    exit 0
endif
