subroutine bcheck3d(tmpl,ul,vl,wl)
  use constants; use mympitools; use mesh; use division
!**********************************************************************
!     benchmark comparizon with Busse et al. [1993]
!     values to be calculated:
!     (1) nusselt number
!     (2) rms velocity
!     (3) temperature gradients at corners
!     (4) temperature at corners of half-depth plane
!     (5) vertical velocity at corners of half-depth plane
!**********************************************************************
  implicit none

  integer :: i,j,k
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: tmpl,ul,vl,wl
!  real(dp),intent(in) :: aspx,aspy!,zp(kax),zc(kax)

  real(dp) :: hfn,tbot,tgd1,tgd2,tgd3,tgd4,tgrd,tsrf,ud1,vd1,wd1,vd2,vrms
  integer :: kp,kc,kprank,kcrank
  real(dp) :: coef1,coef2,tm1,tm2,tm3,tm4,wm1,wm2,wm3,wm4
  real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: &
       twk,uwk,vwk,wwk
  real(dp),dimension(16) :: buffs,buffr

! -----------------------
!     set work arrays
! -----------------------
!
  buffs = 0.e0_dp; buffr = 0.e0_dp
  twk = 0.e0_dp; uwk = 0.e0_dp; vwk = 0.e0_dp; wwk = 0.e0_dp
  do k = ksta,kend
     do j = jsta,jend
        do i = ista,iend
           twk(i,j,k) = tmpl(i,j,k)
           uwk(i,j,k) = ul(i,j,k)
           vwk(i,j,k) = vl(i,j,k)
           wwk(i,j,k) = wl(i,j,k)
        enddo
     enddo
  enddo
  call myshift123l(twk,ista,iend,jsta,jend,ksta,kend)
  call myshift123l(uwk,ista,iend,jsta,jend,ksta,kend)
  call myshift123l(vwk,ista,iend,jsta,jend,ksta,kend)
  call myshift123l(wwk,ista,iend,jsta,jend,ksta,kend)
!
! ==========================
!     (1) nusselt number
! ==========================
!
  tsrf = 0.e0_dp; tbot = 0.e0_dp; tgrd = 0.e0_dp
  if (myrankk == 0) then
     do j = jsta2,jend1
        do i = ista2,iend1
           tbot = tbot + (twk(i,j,1)+twk(i,j,2))*dx(i)*dy(j)
        enddo
     enddo
  endif
  if (myrankk == kprocs-1) then
     do j = jsta2,jend1
        do i = ista2,iend1
           tgrd = tgrd + (twk(i,j,kax-1)-twk(i,j,kax))*dx(i)*dy(j)
           tsrf = tsrf + (twk(i,j,kax-1)+twk(i,j,kax))*dx(i)*dy(j)
        enddo
     enddo
  endif
  tsrf = 0.5e0_dp*tsrf/aspx/aspy
  tbot = 0.5e0_dp*tbot/aspx/aspy
  tgrd = tgrd/aspx/aspy/dzs(kax)
! ------------------------
  buffs(1) = tsrf; buffs(2) = tbot; buffs(3) = tgrd !  hfn = tgrd/(tbot-tsrf)
! ------------------------
!
! ========================
!     (2) rms velocity
! ========================
!
  vrms = 0.e0_dp
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           ud1 = 0.5e0_dp*(uwk(i,j,k)+uwk(i+1,j,k))
           vd1 = 0.5e0_dp*(vwk(i,j,k)+vwk(i,j+1,k))
           wd1 = 0.5e0_dp*(wwk(i,j,k)+wwk(i,j,k+1))
           vd2 = ud1**2+vd1**2+wd1**2
           vrms = vrms + vd2*dx(i)*dy(j)*dz(k)
        enddo
     enddo
  enddo
! ------------------------
  buffs(4) = vrms !  vrms = dsqrt(vrms/aspx/aspy)
! ------------------------
!
! ==========================================================
!     (3) temperature gradient at corners of top surface
! ==========================================================
!
  tgd1 = 0.e0_dp; tgd2 = 0.e0_dp; tgd3 = 0.e0_dp; tgd4 = 0.e0_dp
!
  if (myrank == itable(0,0,kprocs-1)) &
       tgd1 = (twk(2,2,kax-1)-twk(2,2,kax))/dzs(kax)
  if (myrank == itable(iprocs-1,0,kprocs-1)) &
       tgd2 = (twk(iax-1,2,kax-1)-twk(iax-1,2,kax))/dzs(kax)
  if (myrank == itable(iprocs-1,jprocs-1,kprocs-1)) &
       tgd3 = (twk(iax-1,jax-1,kax-1)-twk(iax-1,jax-1,kax))/dzs(kax)
  if (myrank == itable(0,jprocs-1,kprocs-1)) &
       tgd4 = (twk(2,jax-1,kax-1)-twk(2,jax-1,kax))/dzs(kax)
! ------------------------
  buffs(5) = tgd1; buffs(6) = tgd2; buffs(7) = tgd3; buffs(8) = tgd4
! ------------------------
!
! ======================================================
!     (4) temperature at corners of half-depth plane
! ======================================================
!
  tm1 = 0.e0_dp; tm2 = 0.e0_dp; tm3 = 0.e0_dp; tm4 = 0.e0_dp
  do k = 2,kax-1
     if ((zp(k)-0.5e0_dp)*(zp(k-1)-0.5e0_dp) <= 0.e0_dp) then
        kp = k; exit
     endif
  enddo
  coef1 = (0.5e0_dp-zp(kp))/dzs(kp+1)
  coef2 = (zp(kp+1)-0.5e0_dp)/dzs(kp+1)
  do k = 0,kprocs-1
     if ((kstas(k)-kp)*(kends(k)-kp) <= 0) then
        kprank = k; exit
     endif
  enddo
  if (myrank == itable(0,0,kprank)) &
       tm1 = coef1*twk(2,2,kp+1)+coef2*twk(2,2,kp)
  if (myrank == itable(iprocs-1,0,kprank)) &
       tm2 = coef1*twk(iax-1,2,kp+1)+coef2*twk(iax-1,2,kp)
  if (myrank == itable(iprocs-1,jprocs-1,kprank)) &
       tm3 = coef1*twk(iax-1,jax-1,kp+1)+coef2*twk(iax-1,jax-1,kp)
  if (myrank == itable(0,jprocs-1,kprank)) &
       tm4 = coef1*twk(2,jax-1,kp+1)+coef2*twk(2,jax-1,kp)
! -----------------------
  buffs(9) = tm1; buffs(10) = tm2; buffs(11) = tm3; buffs(12) = tm4
! -----------------------
!
! ============================================================
!     (5) vertical velocity at corners of half-depth plane
! ============================================================
!
  wm1 = 0.e0_dp; wm2 = 0.e0_dp; wm3 = 0.e0_dp; wm4 = 0.e0_dp
  do k = 3,kax-1
     if ((zc(k)-0.5e0_dp)*(zc(k-1)-0.5e0_dp) <= 0.e0_dp) then
        kc = k; exit
     endif
  enddo
  coef1 = (0.5e0_dp-zc(kc))/dz(kc)
  coef2 = (zc(kc+1)-0.5e0_dp)/dz(kc)
  do k = 0,kprocs-1
     if ((kstas(k)-kc)*(kends(k)-kc) <= 0) then
        kcrank = k; exit
     endif
  enddo
  if (myrank == itable(0,0,kcrank)) &
       wm1 = coef1*wwk(2,2,kc+1)+coef2*wwk(2,2,kc)
  if (myrank == itable(iprocs-1,0,kcrank)) &
       wm2 = coef1*wwk(iax-1,2,kc+1)+coef2*wwk(iax-1,2,kc)
  if (myrank == itable(iprocs-1,jprocs-1,kcrank)) &
       wm3 = coef1*wwk(iax-1,jax-1,kc+1)+coef2*wwk(iax-1,jax-1,kc)
  if (myrank == itable(0,jprocs-1,kcrank)) &
       wm4 = coef1*wwk(2,jax-1,kc+1)+coef2*wwk(2,jax-1,kc)
! -----------------------
  buffs(13) = wm1; buffs(14) = wm2; buffs(15) = wm3; buffs(16) = wm4
! -----------------------
!
  call mpi_allreduce(buffs,buffr,16,mpi_double_precision,mpi_sum &
       ,mpi_comm_world,ierr)
!
  tsrf = buffr(1); tbot = buffr(2); tgrd = buffr(3); hfn = tgrd/(tbot-tsrf)
  vrms = dsqrt(buffr(4)/aspx/aspy)
  tgd1 = buffr(5); tgd2 = buffr(6); tgd3 = buffr(7); tgd4 = buffr(8)
  tm1 = buffr(9); tm2 = buffr(10); tm3 = buffr(11); tm4 = buffr(12)
  wm1 = buffr(13); wm2 = buffr(14); wm3 = buffr(15); wm4 = buffr(16)
!
  if (myrank == 0) then
     write(24,'("===========================================")')
     write(24,'("  Benchmark Test with Busse et al. [1993]  ")')
     write(24,'("===========================================")')
     write(24,*)"Nu             =",hfn
     write(24,*)"Vrms           =",vrms
     write(24,*)"q1(0, 0, 1),   =",tgd1
     write(24,*)"q2(Lx,0, 1)    =",tgd2
     write(24,*)"q3(Lx,Ly,1)    =",tgd3
     write(24,*)"q4(0, Ly,1)    =",tgd4
     write(24,*)"T1(0, 0, 0.5)  =",tm1
     write(24,*)"T2(Lx,0, 0.5)  =",tm2
     write(24,*)"T3(Lx,Ly,0.5)  =",tm3
     write(24,*)"T4(0, Ly,0.5)  =",tm4
     write(24,*)"Vz1(0, 0, 0.5) =",wm1
     write(24,*)"Vz2(Lx,0, 0.5) =",wm2
     write(24,*)"Vz3(Lx,Ly,0.5) =",wm3
     write(24,*)"Vz4(0, Ly,0.5) =",wm4
  end if
!
  return
end subroutine bcheck3d
