subroutine caleiv(ul,vl,wl,eivl,sivl,viscl)
  use constants; use mympitools; use division
!**********************************************************************
!     calculate 2nd invariants of
!     strain rate/stress tensors
!**********************************************************************
  implicit none

  integer :: i,j,k
  real(dp) :: e11,e22,e33,e12,e13,e23
  real(dp) :: dudx,dudy,dudz,dvdx,dvdy,dvdz,dwdx,dwdy,dwdz
  real(dp) :: dxi1,dxi2,dxi3,dyi1,dyi2,dyi3,dzi1,dzi2,dzi3
  real(dp) :: dum
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend):: ul,vl,wl,eivl,sivl,viscl
  real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: uwk,vwk,wwk

  uwk = 0.e0_dp; vwk = 0.e0_dp; wwk = 0.e0_dp
  do k = ksta,kend
     do j = jsta,jend
        do i = ista,iend
           uwk(i,j,k) = ul(i,j,k)
           vwk(i,j,k) = vl(i,j,k)
           wwk(i,j,k) = wl(i,j,k)
        enddo
     enddo
  enddo

  call myroundshift123l(uwk,ista,iend,jsta,jend,ksta,kend &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)
  call myroundshift123l(vwk,ista,iend,jsta,jend,ksta,kend &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)
  call myroundshift123l(wwk,ista,iend,jsta,jend,ksta,kend &
     ,ista2,iend1,jsta2,jend1,ksta2,kend1)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           dxi1 = 1.e0_dp/(dxs(i)+dxs(i+1))-dxsi(i)
           dxi2 = dxsi(i)-dxsi(i+1)
           dxi3 = -dxi1-dxi2

           dyi1 = 1.e0_dp/(dys(j)+dys(j+1))-dysi(j)
           dyi2 = dysi(j)-dysi(j+1)
           dyi3 = -dyi1-dyi2

           dzi1 = 1.e0_dp/(dzs(k)+dzs(k+1))-dzsi(k)
           dzi2 = dzsi(k)-dzsi(k+1)
           dzi3 = -dzi1-dzi2

           dudx = (uwk(i+1,j,k)-uwk(i,j,k))*dxi(i)

           dudy = .5e0_dp* &
                (dyi1*(uwk(i,j-1,k)+uwk(i+1,j-1,k)) &
                +dyi2*(uwk(i,j,k)+uwk(i+1,j,k)) &
                +dyi3*(uwk(i,j+1,k)+uwk(i+1,j+1,k)))

           dudz = .5e0_dp* &
                (dzi1*(uwk(i,j,k-1)+uwk(i+1,j,k-1)) &
                +dzi2*(uwk(i,j,k)+uwk(i+1,j,k)) &
                +dzi3*(uwk(i,j,k+1)+uwk(i+1,j,k+1)))

           dvdx = .5e0_dp* &
                (dxi1*(vwk(i-1,j,k)+vwk(i-1,j+1,k)) &
                +dxi2*(vwk(i,j,k)+vwk(i,j+1,k)) &
                +dxi3*(vwk(i+1,j,k)+vwk(i+1,j+1,k)))

           dvdy = (vwk(i,j+1,k)-vwk(i,j,k))*dyi(j)

           dvdz = .5e0_dp* &
                (dzi1*(vwk(i,j,k-1)+vwk(i,j+1,k-1)) &
                +dzi2*(vwk(i,j,k)+vwk(i,j+1,k)) &
                +dzi3*(vwk(i,j,k+1)+vwk(i,j+1,k+1)))

           dwdx = .5e0_dp* &
                (dxi1*(wwk(i-1,j,k)+wwk(i-1,j,k+1)) &
                +dxi2*(wwk(i,j,k)+wwk(i,j,k+1)) &
                +dxi3*(wwk(i+1,j,k)+wwk(i+1,j,k+1)))

           dwdy = .5e0_dp* &
                (dyi1*(wwk(i,j-1,k)+wwk(i,j-1,k+1)) &
                +dyi2*(wwk(i,j,k)+wwk(i,j,k+1)) &
                +dyi3*(wwk(i,j+1,k)+wwk(i,j+1,k+1)))

           dwdz = (wwk(i,j,k+1)-wwk(i,j,k))*dzi(k)

           e11 = dudx
           e22 = dvdy
           e33 = dwdz
           e12 = 0.5e0_dp*(dudy+dvdx)
           e13 = 0.5e0_dp*(dudz+dwdx)
           e23 = 0.5e0_dp*(dvdz+dwdy)
           dum = e11**2+e22**2+e33**2+2.e0_dp*(e12**2+e13**2+e23**2)
           eivl(i,j,k) = dsqrt(0.5e0_dp*dum)
        enddo
     enddo
  enddo

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           sivl(i,j,k) = viscl(i,j,k)*eivl(i,j,k)
        enddo
     enddo
  enddo

  return
end subroutine caleiv
