subroutine calvisc(tmp,etl,epl,rat,eiv,visc)
  use constants; use mesh; use division
!**********************************************************************
!     calculate viscosity from temperature and composition
!
!     input
!        tmp: temperature
!        etl: coefficient of temperature on viscosity
!        epl: coefficient of pressure on viscosity
!        eiv: second invariant of strain rate tensor at previous time
!        rat: Rayleigh number defined by viscosity at top surface
!        zp: vertical coodinate
!     output
!        visc: viscosity
!**********************************************************************
  implicit none

  integer :: i,j,k
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend),intent(in) :: tmp,eiv
  real(dp),dimension(ista:iend,jsta:jend,ksta:kend),intent(out) :: visc
  real(dp),intent(in) :: etl,epl,rat
  real(dp) :: zz,visc_creep
  ! yield stress (following by Moresi and Solomatov, 1998)
  ! == rat*(s0+s1*depth)
  real(dp),parameter :: s0=1.e0_dp,s1=0.e0_dp
  real(dp) :: syield,visc_yield

  do k = ksta,kend
     zz = 1.e0_dp-zp(k)
     syield = rat*(s0+s1*zz)
     do j = jsta,jend
        do  i = ista,iend
           visc_creep = exp(-etl*tmp(i,j,k)+epl*zz)
           visc_yield = syield/max(1.e-30_dp,eiv(i,j,k))
           visc(i,j,k) = min(visc_creep,visc_yield)
           if (visc_yield < visc_creep) &
                write(*,*)"Yielded",i,j,k,visc_yield,visc_creep
        enddo
     enddo
  enddo

  return
end subroutine calvisc
