subroutine fileout10(tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl,file10,iflag)
  use constants; use mympitools; use division; use mesh
!**********************************************************************
!     output data to file10
!     iflag=1 when each process writes *own* data for *all* arrays
!     iflag=2 when root process writes *all* data for *each* array
!     otherwise    root process writes *all* data for *all* arrays
!**********************************************************************
  implicit none

  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: &
       tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl
  integer,intent(in) :: iflag
  character(*) :: file10

  select case (iflag)
  case (1)
     open(10,file=file10,form="unformatted")
     call fileout10_1(tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl)
     close(10)
  case (2)
     call fileout10_2(tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl,file10)
  case default
     open(10,file=file10,form="unformatted")
     call fileout10_0(tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl)
     close(10)
  end select

  return

contains

! ---------------------------------------------------------------------

  subroutine fileout10_0(tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl)
    implicit none
    integer :: i,j,k,ijk
    real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: &
         tmpl,schl,ul,vl,wl,pl,sivl,eivl,viscl
    integer,parameter :: nvals=9
    real(dp),dimension(:,:,:,:),allocatable :: outdat

    ! write(10)tmp,u,v,w,sch,siv,eiv,visc
    allocate(outdat(iax,jax,kax,nvals))
    call mygather123lg(tmpl, outdat(1,1,1,1),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    call mygather123lg(ul,   outdat(1,1,1,2),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    call mygather123lg(vl,   outdat(1,1,1,3),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    call mygather123lg(wl,   outdat(1,1,1,4),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    call mygather123lg(pl,   outdat(1,1,1,5),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    call mygather123lg(schl, outdat(1,1,1,6),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    call mygather123lg(sivl, outdat(1,1,1,7),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    call mygather123lg(eivl, outdat(1,1,1,8),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    call mygather123lg(viscl,outdat(1,1,1,9),ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    if (myrank.eq.0)write(10)outdat
    deallocate(outdat)

    return
  end subroutine fileout10_0

! ---------------------------------------------------------------------

  subroutine fileout10_1(tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl)
!**********************************************************************
!**********************************************************************
    implicit none

    real(dp),dimension(ista:iend,jsta:jend,ksta:kend),intent(in) :: &
         tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl

    write(10)ista,iend,jsta,jend,ksta,kend &
         ,tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl

    return
  end subroutine fileout10_1

! ---------------------------------------------------------------------

  subroutine fileout10_2(tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl,file10)
!**********************************************************************
!**********************************************************************
    implicit none

    integer :: i,j,k,ij
    real(dp),dimension(ista:iend,jsta:jend,ksta:kend),intent(in) :: &
         tmpl,schl,ul,vl,wl,pl,sivl,eivl,viscl
    integer,parameter :: nvalues=9
    real(dp),dimension(:,:,:),allocatable :: outdat
    character(*) :: file10
    character(80) :: output_file
    character(1),dimension(nvalues) :: head
!
    allocate(outdat(iax,jax,kax))
    head = (/ "t","u","v","w","p","q","s","e","h" /)

    ! write(10)tmp,u,v,w,sch,siv,eiv,visc

    call mygather123lg(tmpl,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "t"//file10
    open(31,file=output_file,form="unformatted")
    write(31)outdat

    call mygather123lg(ul,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "u"//file10
    open(32,file=output_file,form="unformatted")
    write(32)outdat

    call mygather123lg(vl,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "v"//file10
    open(33,file=output_file,form="unformatted")
    write(33)outdat

    call mygather123lg(wl,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "w"//file10
    open(34,file=output_file,form="unformatted")
    write(34)outdat

    call mygather123lg(pl,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "p"//file10
    open(35,file=output_file,form="unformatted")
    write(35)outdat

    call mygather123lg(schl,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "q"//file10
    open(36,file=output_file,form="unformatted")
    write(36)outdat

    call mygather123lg(sivl,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "s"//file10
    open(37,file=output_file,form="unformatted")
    write(37)outdat

    call mygather123lg(eivl,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "e"//file10
    open(38,file=output_file,form="unformatted")
    write(38)outdat

    call mygather123lg(viscl,outdat,ista,iend,jsta,jend,ksta,kend &
         ,iax,jax,kax)
    output_file = "h"//file10
    open(39,file=output_file,form="unformatted")
    write(39)outdat

    close(31); close(32); close(33); close(34);close(35); close(36)
    close(37); close(38); close(39)
    deallocate(outdat)
!
    return
  end subroutine fileout10_2

end subroutine fileout10
