subroutine init(tmp,q,tsurf,tbot,tamp,siv,eiv,visc,u,v,w,p,rat,&
     icont,iflag,file9)
  use constants; use mympitools; use mesh; use division
!**********************************************************************
!     set initial conditions
!
!     icont: = 0 if new run
!            = 1 if successive run
!     ivlt: = 1 if free-slip top surface
!           = 2 if fixed velocity top surface
!     input
!     output
!        tmp: temperature field
!        q: internal heating rate
!        tamp: reference temperature
!        bvlt: -2*(surface velocity)*dz
!**********************************************************************
  implicit none

  integer :: i,j,k
  integer :: icont,ivlt

  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: &
       tmp,q,visc,eiv,siv,u,v,w,p
  real(dp) :: bvlt(iax+1)
  real(dp) :: rat
  real(dp) :: tsurf,tbot,tamp
  real(dp) :: cellx,celly,dtbl,zz
  real(dp) :: t1,t2,tm,tza
  character(80) :: file9
  integer :: iflag
  real(dp),dimension(kax) :: tbg,pbg

! ====================================
  if (icont == 0) then ! for new run
! ====================================

! -------------------------------
!     set initial temperature
! -------------------------------

     ! background temerature
     tm = 0.5e0_dp*(tsurf+tbot); dtbl = .5e0_dp ! give conductive state
     do k = 2,kax-1
        zz = zp(k)
        t1 = tsurf+(1.d0-zz)*(tm-tsurf)/dtbl
        t2 = dmax1(tbot+zz*(tm-tbot)/dtbl,tm)
        tbg(k) = dmin1(t1,t2)
     end do
     ! add perturbation
     cellx = 1.e0_dp; celly = 1.e0_dp
     do k = ksta2,kend1
        tza = tamp*dsin(pi*zp(k))
        do j = jsta2,jend1
           do i = ista2,iend1
#ifdef BOX2D
              tmp(i,j,k) = tbg(k)+tza*dcos(pi*cellx*xp(i)/aspx)
#else
              tmp(i,j,k) = tbg(k)+tza*(dcos(pi*cellx*xp(i)/aspx) &
                   +dcos(pi*celly*yp(j)/aspy))
#endif
           enddo
        enddo
     enddo

     if (myrankk == 0) then
        tmp(ista2:iend1,jsta2:jend1,1) = &
             2.e0_dp*tbot - tmp(ista2:iend1,jsta2:jend1,2)
     endif
     if (myrankk == kprocs-1) then
        tmp(ista2:iend1,jsta2:jend1,kax) = 2.e0_dp*tsurf &
             - tmp(ista2:iend1,jsta2:jend1,kax-1)
     endif
     if (myranki == 0) then
        tmp(1,jsta2:jend1,ksta:kend) = tmp(2,jsta2:jend1,ksta:kend)
     endif
     if (myranki == iprocs-1) then
        tmp(iax,jsta2:jend1,ksta:kend) = tmp(iax-1,jsta2:jend1,ksta:kend)
     endif
     if (myrankj == 0) then
        tmp(ista:iend,1,ksta:kend) = tmp(ista:iend,2,ksta:kend)
     endif
     if (myrankj == jprocs-1) then
        tmp(ista:iend,jax,ksta:kend) = tmp(ista:iend,jax-1,ksta:kend)
     endif

! -----------------------------------
!     set initial pressure fields
! -----------------------------------
     pbg(2) = 0.e0_dp
     do k = 3,kax-1
        pbg(k) = pbg(k-1)+0.5e0*rat*(tbg(k)*dz(k)+tbg(k-1)*dz(k-1))
     end do
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              p(i,j,k) = pbg(k)
           end do
        end do
     end do

! ----------------------------------------
!     set initial heat_production_rate
! ----------------------------------------

     q = 0.e0_dp

! -----------------------------------
!     set initial velocity fields
! -----------------------------------

     u = 0.e0_dp; v = 0.e0_dp; w = 0.e0_dp

     tamp = tbot - tsurf

! ===========================================
  else if (icont == 1) then ! restaring run
! ===========================================

     call filein(tmp,u,v,w,p,q,siv,eiv,visc,file9,iflag)
  endif

  return
end subroutine init
