program main123
  use constants; use mympitools; use mesh; use division; use hdesim3
!**********************************************************************
!     mantle convection in three-dimensional rectangular box
!     hydrodynamic equations are solved by SIMPLER method
!     thermal equation are solved by explicit or implicit scheme
!     heat flux is defined by power-law scheme
!     originally written by M. Kameyama on 20030620
!     final version by M. Kameyama on 20040310
!**********************************************************************
  implicit none
  integer :: i,j,k,l

  real(dp),dimension(:,:,:),allocatable :: &
       ul,vl,wl,pl,viscl,schl,tmpl,sivl,eivl,buoyl,sivlo,eivlo,visclo
  real(dp) :: rat,tsurf,tbot,tamp,qbot,vboti!,aspx,aspy
  integer :: nt,ntime,maxtime,iplot,iprnt,nstep
  real(dp) :: time,dt
  integer :: ivlt,ivlb,icont,ipt,ivis
  real(dp) :: etl,etn,epl,epn
  real(dp) :: relu,relv,relw,relp,crmax
!
!     choose characteristic of files
!     iflag=1 when each process reads/writes *own* data for *all* arrays
!     iflag=2 when each process reads/writes *all* data for *each* array
!     otherwise    each process reads/writes *all* data for *all* arrays
!
  character(80) :: ipf,opf,flog,file9h,file10h,file11h,file20
  character(87) :: file9,file10,file11
  integer :: iflag9,iflag10,iflag11

!     for parallelization
!
!  integer :: ista,iend,jsta,jend,ksta,kend
!
!     choose preconditioner for CG solvers in SIMPLER method
!     isolver = 0 for None
!             = 1 for Localized ILU(0) decomposition
!             = 2 for Symmetric SOR
!             = 3 for Point-Jacobi (diagonal scaling)
  integer :: isolvers(6),isolver

! ----------------------------------------
!     read names of files to be opened
! ----------------------------------------

  open(23,file="finput")
  read(23,'(a80)')ipf
  read(23,'(a80)')opf
  read(23,'(a80)')flog
  read(23,'(a80)')file9h
  read(23,'(a80)')file10h
  read(23,'(a80)')file11h
  read(23,'(a80)')file20
  close(23)

! --------------------
!     read controls
! --------------------
!
  open(23,file="run_controls")
  read(23,*)iflag9,iflag10,iflag11
  read(23,*)isolvers
  close(23)

! ----------------------
  call mympiinit
! ----------------------
!
! ------------------------------------------
!     define names of files to be opened
! ------------------------------------------
!
  if (iflag9 == 1) then
     file9 = distributed_file_name(file9h,myrank)
  else
     file9 = file9h
  endif
  if (iflag10 == 1) then
     file10 = distributed_file_name(file10h,myrank)
  else
     file10 = file10h
  endif
  if (iflag11 == 1) then
     file11 = distributed_file_name(file11h,myrank)
  else
     file11 = file11h
  endif

! --------------------
!     split domain
! --------------------

  call proc_table
  call para_range123(1,iax,1,jax,1,kax)
  call divide_region
!
! -----------------------------
!     allocate local arrays
! -----------------------------
!
  allocate &
       (ul(ista:iend,jsta:jend,ksta:kend) &
       ,vl(ista:iend,jsta:jend,ksta:kend) &
       ,wl(ista:iend,jsta:jend,ksta:kend) &
       ,pl(ista:iend,jsta:jend,ksta:kend) &
       ,viscl(ista:iend,jsta:jend,ksta:kend) &
       ,schl(ista:iend,jsta:jend,ksta:kend) &
       ,tmpl(ista:iend,jsta:jend,ksta:kend) &
       ,sivl(ista:iend,jsta:jend,ksta:kend) &
       ,eivl(ista:iend,jsta:jend,ksta:kend) &
       ,buoyl(ista:iend,jsta:jend,ksta:kend) &
       ,sivlo(ista:iend,jsta:jend,ksta:kend) &
       ,eivlo(ista:iend,jsta:jend,ksta:kend) &
       ,visclo(ista:iend,jsta:jend,ksta:kend))

! ----------------------
!     initialize run
! ----------------------

  if (myrank == 0) open(24,file=flog)
  call rdata(maxtime,nstep,iplot,iprnt,time,ivlt,ivlb,icont, &
       ivis,ipt,etl,etn,epl,epn,rat,tsurf,tbot,tamp,vboti, &
       ipf,relu,relv,relw,relp,crmax)
  call dfmesh
  call mconst
  if (myrank.eq.0) then
     write(24,'(8a10)') &
          "dxmax","dxmin","dymax","dymin","dzmax","dzmin","dxyzmax","dxyzmin"
     write(24,'(8f10.7)')dxmax,dxmin,dymax,dymin,dzmax,dzmin,dxyzmax,dxyzmin
  endif
  call init(tmpl,schl,tsurf,tbot,tamp,sivl,eivl,viscl,ul,vl,wl,pl,rat, &
       icont,iflag9,file9)

  if (myrank == 0) then
     write(*,'(" === Domain Partitioning ===")')
     write(*,'(7a8)')"rank","ista","iend","jsta","jend","ksta","kend"
     do k = 0,kprocs-1
        do j = 0,jprocs-1
           do i = 0,iprocs-1
              write(*,'(7i8)')itable(i,j,k),istas(i),iends(i) &
                   ,jstas(j),jends(j),kstas(k),kends(k)
           end do
        end do
     end do
  end if

! ==========================
!     begin time marching
! ==========================

  open(11,file=file11)
  if (myrank == 0) open(20,file=file20)
  do nt = 1,maxtime
     ntime = nt + nstep

     eivlo = eivl; sivlo = sivl; visclo = viscl

! ----------------------------------------------
!     calculate buoyancy term at mesh points
! ----------------------------------------------

     call buoyant(tmpl,rat,buoyl)

! -----------------------------------
!     solve hydrodynamic equation
! -----------------------------------

     call calvisc(tmpl,etl,epl,rat,eivl,viscl)
     call hdesim3main(ul,vl,wl,pl,viscl,buoyl,ivlt,ivlb, &
          ivis,relu,relv,relw,relp,isolvers)
     call caleiv(ul,vl,wl,eivl,sivl,viscl)

! --------------------------
!     evolve temperature
! --------------------------

     call tstep(ul,vl,wl,sivl,eivl,crmax,dt)
     call htrns3(tmpl,ul,vl,wl,schl,tsurf,tbot,qbot,vboti,dt,crmax,isolvers(6))
    time = time + dt

!     for printing
     if (mod(nt,iprnt)  ==  0) then
        call plot(tmpl,tsurf,tbot,qbot,ul,vl,wl,eivl,sivl,time,ntime)
     endif

!     for drawing

     if (iplot /= 0) then
        if (mod(nt,iplot) == 0) then
           call fileout11(ntime,time,tmpl,ul,vl,wl,schl,sivl,eivl,viscl, &
                iflag11)
        endif
     endif

  enddo

! =========================
!     end time marching
! =========================

! ----------------------
!     output results
! ----------------------

  if (mod(nt-1,iprnt) /= 0) then
     call plot(tmpl,tsurf,tbot,qbot,ul,vl,wl,eivl,sivl,time,ntime-1)
  endif

!     for a next run

  call fileout10(tmpl,ul,vl,wl,pl,schl,sivl,eivl,viscl,file10,iflag10)

  call wdata(maxtime,nstep,iplot,iprnt,time,ivlt,ivlb,icont, &
       ivis,ipt,etl,etn,epl,epn,rat,tsurf,tbot,tamp,vboti &
       ,aspx,aspy,opf,relu,relv,relw,relp,crmax)

!     benckmark comparison

  call bcheck3d(tmpl,ul,vl,wl)

! --------------------------
  call mympifinalize
! --------------------------

  stop
end program main123
