module mylinsol
  use constants; use mympitools
  implicit none
  private
  public :: bcgbase123pl, cgbase123pl, jacobi123p, sor2c123p

contains
! =============================================================================
subroutine bcgbase123pl(amt,bvc,xvc,eitr,nitr,istat &
     ,amtprec,d,ista2,iend1,jsta2,jend1,ksta2,kend1 &
     ,isolver,iprec,prec)

  integer,intent(in) :: isolver
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
       :: amt,amtprec
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
       :: bvc,xvc,d
  real(dp),intent(inout) :: eitr
  integer,intent(inout) :: nitr,istat,iprec
  character(*),intent(out) :: prec

  select case (isolver)
  case default
     prec = "NoPC"
     call purebcg123pl(amt,bvc,xvc,eitr,nitr,istat &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)
  case (1)
     prec = "LILU"
     call ilubcg123pl(amt,bvc,xvc,eitr,nitr,istat,iprec &
          ,amtprec,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
!  case (2)
!     call ssorcg123pl(amt,bvc,xvc,eitr,nitr,istat,iprec &
!          ,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
  case (3)
     prec = "PJAC"
     call pjbcg123pl(amt,bvc,xvc,eitr,nitr,istat,iprec &
          ,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
  end select

  return
end subroutine bcgbase123pl
! =============================================================================
subroutine cgbase123pl(amt,bvc,xvc,eitr,nitr,istat &
     ,amtprec,d,ista2,iend1,jsta2,jend1,ksta2,kend1 &
     ,isolver,iprec,prec)

  integer,intent(in) :: isolver
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) &
       :: amt,amtprec
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
       :: bvc,xvc,d
  real(dp),intent(inout) :: eitr
  integer,intent(inout) :: nitr,istat,iprec
  character(*),intent(out) :: prec

  select case (isolver)
  case default
     prec = "NoPC"
     call purecg123pl(amt,bvc,xvc,eitr,nitr,istat &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)
  case (1)
     prec = "LILU"
     call iccg123pl(amt,bvc,xvc,eitr,nitr,istat,iprec &
          ,amtprec,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
  case (2)
     prec = "SSOR"
     call ssorcg123pl(amt,bvc,xvc,eitr,nitr,istat,iprec &
          ,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
  case (3)
     prec = "PJAC"
     call pjcg123pl(amt,bvc,xvc,eitr,nitr,istat,iprec &
          ,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
  end select

  return
end subroutine cgbase123pl
! =============================================================================
subroutine purebcg123pl(a,b,x,eps,nitr,ier &
     ,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
! Bi-conjugate gradient method with localized ILU(0) preconditioning
!
!     copyright: Tsutomu Oguni       Feb. 1 1993         ver. 2
!     modified by masanori kameyama
!     on 20030106; limited to 3-D finite difference equation
!        20030117; parallelized by localized ILU(0) preconditioning
!        20030205: domain partitioning by all three directions
!        20030403: preconditioner should be defined/decomposed
!                  before calling this subroutine
!        20030408: shift-type communication for matrix should be done
!                  before calling this subroutine
!*********************************************************************

  integer :: i,j,k,ij
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  integer :: ier,itr,nitr,maxitr
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: b,x
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,2) :: p,q,r
  real(dp) :: eps,alpha,beta,c1,c2,c3,res,x1,x2

  call myshift4a123l(a,ista2,iend1,jsta2,jend1,ksta2,kend1)

! ======================
!     initialization
! ======================
  p = 0.e0_dp; q = 0.e0_dp; r = 0.e0_dp

!     r=M^-1*(b-A*x)
  call naxsub123pl(a,x,q(ista2-1,jsta2-1,ksta2-1,1) &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k,1) = b(i,j,k)-q(i,j,k,1)
        enddo
     enddo
  enddo

!     define other vectors
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k,2) = r(i,j,k,1)
           p(i,j,k,1) = r(i,j,k,1)
           p(i,j,k,2) = r(i,j,k,1)
        enddo
     enddo
  enddo

!     c1 = (r,r)
  call vnorm123pl(r(ista2-1,jsta2-1,ksta2-1,1),c1 &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)

! +++++++++++++++++++++++++++++++++++
!     ILUBCG iteration loop start
! +++++++++++++++++++++++++++++++++++
  maxitr = nitr; ier = 1
  bcgloop: do itr = 1,maxitr

!     q~=At*p~
     call natsub123pl(a,p(ista2-1,jsta2-1,ksta2-1,2) &
          ,q(ista2-1,jsta2-1,ksta2-1,2) &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     q=A*p
     call naxsub123pl(a,p(ista2-1,jsta2-1,ksta2-1,1) &
          ,q(ista2-1,jsta2-1,ksta2-1,1) &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     c2=(q,p~)
     call vdotv123pl(q(ista2-1,jsta2-1,ksta2-1,1) &
          ,p(ista2-1,jsta2-1,ksta2-1,2),c2 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

     if (c2 == 0.e0_dp) then
        ier = 5; nitr = itr
        exit bcgloop
     endif

     call vnorm123pl(x,x1,ista2,iend1,jsta2,jend1,ksta2,kend1)
     call vnorm123pl(p(ista2-1,jsta2-1,ksta2-1,1),x2 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)
     alpha = c1 / c2
     x2 = x2*alpha**2

     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              x(i,j,k) = x(i,j,k)+alpha*p(i,j,k,1)     ! x=x+alpha*p
              r(i,j,k,1) = r(i,j,k,1)-alpha*q(i,j,k,1) ! r=r-alpha*q
              r(i,j,k,2) = r(i,j,k,2)-alpha*q(i,j,k,2) ! r~=r~-alpha*q~
           enddo
        enddo
     enddo

!     c3=(r,r~)
     call vdotv123pl(r(ista2-1,jsta2-1,ksta2-1,1) &
          ,r(ista2-1,jsta2-1,ksta2-1,2),c3 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     convergency check
!     1. correction of x is small enough
     if (x1 /= 0.e0_dp) then
        res = dsqrt(x2 / x1)
        if (res < eps) then
           nitr = itr; ier = 0
           exit bcgloop
        endif
     endif

!     2. residue is 0
     if (c1 == 0.e0_dp) then
        ier = 4; nitr = itr
        exit bcgloop
     endif

!     update p and p~
     beta = c3 / c1
     c1 = c3
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              p(i,j,k,1) = r(i,j,k,1)+beta*p(i,j,k,1)
              p(i,j,k,2) = r(i,j,k,2)+beta*p(i,j,k,2)
           enddo
        enddo
     enddo
! +++++++++++++++++++++++++++++++++
!     ILUBCG iteration loop end
! +++++++++++++++++++++++++++++++++
  enddo bcgloop
  eps = res

  return
end subroutine purebcg123pl
! =============================================================================
subroutine ilubcg123pl(a,b,x,eps,nitr,ier,iprec &
     ,am,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
! Bi-conjugate gradient method with localized ILU(0) preconditioning
!
!     copyright: Tsutomu Oguni       Feb. 1 1993         ver. 2
!     modified by masanori kameyama
!     on 20030106; limited to 3-D finite difference equation
!        20030117; parallelized by localized ILU(0) preconditioning
!        20030205: domain partitioning by all three directions
!        20030403: preconditioner should be defined/decomposed
!                  before calling this subroutine
!        20030408: shift-type communication for matrix should be done
!                  before calling this subroutine
!*********************************************************************

  integer :: i,j,k,ij
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  integer :: ier,itr,nitr,maxitr,iprec
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a,am
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: b,x,d,wrk
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,2) :: p,q,r
  real(dp) :: eps,alpha,beta,c1,c2,c3,res,x1,x2

! =======================
!     preconditioning
!     matrix shift + localized ILU(0) decomposition
! =======================
  if (iprec == 0) then
     call myshift4a123l(a,ista2,iend1,jsta2,jend1,ksta2,kend1)
     call liludcmp123pl(a,am,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
     iprec = 1
  endif

! ======================
!     initialization
! ======================
  p = 0.e0_dp; q = 0.e0_dp; r = 0.e0_dp; wrk = 0.e0_dp

!     r=M^-1*(b-A*x)
  call naxsub123pl(a,x,q(ista2-1,jsta2-1,ksta2-1,1) &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           wrk(i,j,k) = b(i,j,k)-q(i,j,k,1)
        enddo
     enddo
  enddo
  call naxsl123pl(am,d,wrk,r(ista2-1,jsta2-1,ksta2-1,1) &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     define other vectors
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k,2) = r(i,j,k,1)
           p(i,j,k,1) = r(i,j,k,1)
           p(i,j,k,2) = r(i,j,k,1)
        enddo
     enddo
  enddo

!     c1 = (r,r)
  call vnorm123pl(r(ista2-1,jsta2-1,ksta2-1,1),c1 &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)

! +++++++++++++++++++++++++++++++++++
!     ILUBCG iteration loop start
! +++++++++++++++++++++++++++++++++++
  maxitr = nitr; ier = 1
  bcgloop: do itr = 1,maxitr

!     q~=(M^-1*A)t*p~
     call natsl123pl(am,d,p(ista2-1,jsta2-1,ksta2-1,2) &
          ,q(ista2-1,jsta2-1,ksta2-1,1) &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)
     call natsub123pl(a,q(ista2-1,jsta2-1,ksta2-1,1) &
          ,q(ista2-1,jsta2-1,ksta2-1,2) &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     q=(M^-1)*A*p
     call naxsub123pl(a,p(ista2-1,jsta2-1,ksta2-1,1),wrk &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)
     call naxsl123pl(am,d,wrk,q(ista2-1,jsta2-1,ksta2-1,1) &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     c2=(q,p~)
     call vdotv123pl(q(ista2-1,jsta2-1,ksta2-1,1) &
          ,p(ista2-1,jsta2-1,ksta2-1,2),c2 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

     if (c2 == 0.e0_dp) then
        ier = 5; nitr = itr
        exit bcgloop
     endif

     call vnorm123pl(x,x1,ista2,iend1,jsta2,jend1,ksta2,kend1)
     call vnorm123pl(p(ista2-1,jsta2-1,ksta2-1,1),x2 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)
     alpha = c1/c2
     x2 = x2*alpha**2

!     x=x+alpha*p
!     r=r-alpha*q
!     r~=r~-alpha*q~
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              x(i,j,k) = x(i,j,k)+alpha*p(i,j,k,1)
              r(i,j,k,1) = r(i,j,k,1)-alpha*q(i,j,k,1)
              r(i,j,k,2) = r(i,j,k,2)-alpha*q(i,j,k,2)
           enddo
        enddo
     enddo

!     c3=(r,r~)
     call vdotv123pl(r(ista2-1,jsta2-1,ksta2-1,1) &
          ,r(ista2-1,jsta2-1,ksta2-1,2),c3 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     convergency check
!     1. correction of x is small enough
     if (x1 /= 0.e0_dp) then
        res = dsqrt(x2 / x1)
        if (res <= eps) then
           nitr = itr; ier = 0
           exit bcgloop
        endif
     endif
!     2. residue is 0
     if (c1 == 0.e0_dp) then
        ier = 4; nitr = itr
        exit bcgloop
     endif

!     update p and p~
     beta = c3 / c1
     c1 = c3
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              p(i,j,k,1) = r(i,j,k,1)+beta*p(i,j,k,1)
              p(i,j,k,2) = r(i,j,k,2)+beta*p(i,j,k,2)
           enddo
        enddo
     enddo
! +++++++++++++++++++++++++++++++++
!     ILUBCG iteration loop end
! +++++++++++++++++++++++++++++++++
  enddo bcgloop
  eps = res

  return
end subroutine ilubcg123pl
! =============================================================================
subroutine pjbcg123pl(a,b,x,eps,nitr,ier,iprec &
     ,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
! Bi-conjugate gradient method with Point Jacobi preconditioning
!
!     copyright: Tsutomu Oguni       Feb. 1 1993         ver. 2
!     modified by masanori kameyama
!     on 20030604
!*********************************************************************

  integer :: i,j,k,ij
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  integer :: ier,itr,nitr,maxitr,iprec
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: b,x,wrk,d
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,2) :: p,q,r
  real(dp) :: eps
  real(dp) :: alpha,beta,c1,c2,c3,res,x1,x2

! =======================
!     preconditioning
!     matrix shift and diagonal scaling
! =======================
  if (iprec == 0) then
     call myshift4a123l(a,ista2,iend1,jsta2,jend1,ksta2,kend1)
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              d(i,j,k) = 1.e0_dp/a(i,j,k,4)
           end do
        end do
     end do
     iprec = 1
  endif

! ======================
!     initialization
! ======================
  p = 0.e0_dp; q = 0.e0_dp; r = 0.e0_dp; wrk = 0.e0_dp

!     r=M^-1*(b-A*x)
  call naxsub123pl(a,x,q(ista2-1,jsta2-1,ksta2-1,1) &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           wrk(i,j,k) = b(i,j,k)-q(i,j,k,1)
        enddo
     enddo
  enddo
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k,1) = d(i,j,k)*wrk(i,j,k)
        end do
     end do
  end do

!     define other vectors
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k,2) = r(i,j,k,1)
           p(i,j,k,1) = r(i,j,k,1)
           p(i,j,k,2) = r(i,j,k,1)
        enddo
     enddo
  enddo

!     c1 = (r,r)
  call vnorm123pl(r(ista2-1,jsta2-1,ksta2-1,1),c1 &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)

! +++++++++++++++++++++++++++++++++++
!     ILUBCG iteration loop start
! +++++++++++++++++++++++++++++++++++
  maxitr = nitr; ier = 1
  bcgloop: do itr = 1,maxitr

!     q~=(M^-1*A)t*p~
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              q(i,j,k,1) = d(i,j,k)*p(i,j,k,2)
           end do
        end do
     end do
     call natsub123pl(a,q(ista2-1,jsta2-1,ksta2-1,1) &
          ,q(ista2-1,jsta2-1,ksta2-1,2) &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     q=(M^-1)*A*p
     call naxsub123pl(a,p(ista2-1,jsta2-1,ksta2-1,1),wrk &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              q(i,j,k,1) = d(i,j,k)*wrk(i,j,k)
           end do
        end do
     end do

!     c2=(q,p~)
     call vdotv123pl(q(ista2-1,jsta2-1,ksta2-1,1) &
          ,p(ista2-1,jsta2-1,ksta2-1,2),c2 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

     if (c2 == 0.e0_dp) then
        ier = 5; nitr = itr
        exit bcgloop
     endif

     call vnorm123pl(x,x1,ista2,iend1,jsta2,jend1,ksta2,kend1)
     call vnorm123pl(p(ista2-1,jsta2-1,ksta2-1,1),x2 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)
     alpha = c1 / c2
     x2 = x2*alpha**2

     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              x(i,j,k) = x(i,j,k)+alpha*p(i,j,k,1)     ! x=x+alpha*p
              r(i,j,k,1) = r(i,j,k,1)-alpha*q(i,j,k,1) ! r=r-alpha*q
              r(i,j,k,2) = r(i,j,k,2)-alpha*q(i,j,k,2) ! r~=r~-alpha*q~
           enddo
        enddo
     enddo

!     c3=(r,r~)
     call vdotv123pl(r(ista2-1,jsta2-1,ksta2-1,1) &
          ,r(ista2-1,jsta2-1,ksta2-1,2),c3 &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1)

!     convergency check
!     1. correction of x is small enough
     if (x1 /= 0.e0_dp) then
        res = dsqrt(x2 / x1)
        if (res .le. eps) then
           nitr = itr; ier = 0
           exit bcgloop
        endif
     endif

!     2. residue is 0
     if (c1 == 0.e0_dp) then
        ier = 4; nitr = itr
        exit bcgloop
     endif

!     update p and p~
     beta = c3 / c1
     c1 = c3
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              p(i,j,k,1) = r(i,j,k,1)+beta*p(i,j,k,1)
              p(i,j,k,2) = r(i,j,k,2)+beta*p(i,j,k,2)
           enddo
        enddo
     enddo
! +++++++++++++++++++++++++++++++++
!     ILUBCG iteration loop end
! +++++++++++++++++++++++++++++++++
  enddo bcgloop
  eps = res

  return
end subroutine pjbcg123pl
! =============================================================================
subroutine purecg123pl(a,b,x,eps,nitr,istat &
     ,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     conjugate gradient method
!     solve 3-D Poisson equation by FDM
!
!     ver.0 on 20030604
!     ver.1 on 20030625: inline expansion of naxsub,vdotv,vnorm
!*********************************************************************

  integer :: i,j,k,ncg,maxitr,ij,nitr
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: b,x,r,p,ap
  real(dp) :: alpha,b2,beta,eps,eps2,rur0,rur1,pap,res2
  integer :: istat

  maxitr = nitr; eps2 = eps**2

! ======================
!     initialization
! ======================
  p = 0.e0_dp; r = 0.e0_dp; ap = 0.e0_dp

  call naxsub123pl(a,x,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k) = b(i,j,k)-ap(i,j,k)
        end do
     end do
  end do

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           p(i,j,k) = r(i,j,k)
        end do
     end do
  end do
  call vnorm123pl(r,rur0,ista2,iend1,jsta2,jend1,ksta2,kend1)
  call vnorm123pl(b,b2,ista2,iend1,jsta2,jend1,ksta2,kend1)

! ============================
!     iterative loop start
! ============================
  istat = 1
  cgloop: do ncg = 1,maxitr
!cdir iexpand
     call naxsub123pl(a,p,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)
!cdir iexpand
     call vdotv123pl(p,ap,pap,ista2,iend1,jsta2,jend1,ksta2,kend1)

     alpha = rur0/pap
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              x(i,j,k) = x(i,j,k)+alpha*p(i,j,k)
              r(i,j,k) = r(i,j,k)-alpha*ap(i,j,k)
           end do
        end do
     end do
!cdir iexpand
     call vnorm123pl(r,res2,ista2,iend1,jsta2,jend1,ksta2,kend1)

     eps = dsqrt(res2/b2)

     if (res2/b2.le.eps2) then ! converged
        nitr = ncg; istat = 0
        exit cgloop
     end if

     !call vnorm123pl(r,rur1,ista2,iend1,jsta2,jend1,ksta2,kend1)
     rur1 = res2
     beta = rur1/rur0
     rur0 = rur1

     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              p(i,j,k) = r(i,j,k)+beta*p(i,j,k)
           end do
        end do
     end do
! ==========================
!     iterative loop end
! ==========================
  end do cgloop

  return
end subroutine purecg123pl
! =============================================================================
subroutine iccg123pl(a,b,x,eps,nitr,istat,iprec &
     ,am,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     conjugate gradient method with ILU(0) preconditioning
!     solve 3-D Poisson equation by FDM
!
!     ver.0 on 20030120; localized ILU(0) preconditioning
!                        is implemented
!     ver.1 on 20030205: domain partitioning in all 3 dimensions
!     ver.2 on 20030319: array reduction
!     ver.3 on 20030403: preconditioner should be defined/decomposed
!                        before calling this subroutine
!     ver.4 on 20030625: inline expansion of naxsub,vdotv,vnorm
!*********************************************************************

  integer :: i,j,k,ncg,maxitr,ij,nitr
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a,am
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: b,x,r,p,ru,ap,d
  real(dp) :: alpha,b2,beta,eps,eps2
  real(dp) :: rur0,rur1,pap,res2
  integer :: istat,iprec

  maxitr = nitr; eps2 = eps**2

! ===============================
!     LILU(0) preconditioning
! ===============================
  if (iprec == 0) then
     call liludcmp123pl(a,am,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
     iprec = 1
  end if

! ======================
!     initialization
! ======================
  p = 0.e0_dp; r = 0.e0_dp; ru = 0.e0_dp; ap = 0.e0_dp

  call naxsub123pl(a,x,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k) = b(i,j,k)-ap(i,j,k)
        end do
     end do
  end do

  call naxsl123pl(am,d,r,ru,ista2,iend1,jsta2,jend1,ksta2,kend1)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           p(i,j,k) = ru(i,j,k)
        end do
     end do
  end do
  call vdotv123pl(r,ru,rur0,ista2,iend1,jsta2,jend1,ksta2,kend1)

  call vnorm123pl(b,b2,ista2,iend1,jsta2,jend1,ksta2,kend1)

! ============================
!     iterative loop start
! ============================
  istat = 1
  cgloop: do ncg = 1,maxitr
!cdir iexpand
     call naxsub123pl(a,p,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)
!cdir iexpand
     call vdotv123pl(p,ap,pap,ista2,iend1,jsta2,jend1,ksta2,kend1)

     alpha = rur0/pap
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              x(i,j,k) = x(i,j,k)+alpha*p(i,j,k)
              r(i,j,k) = r(i,j,k)-alpha*ap(i,j,k)
           end do
        end do
     end do

!cdir iexpand
     call vnorm123pl(r,res2,ista2,iend1,jsta2,jend1,ksta2,kend1)

     eps = dsqrt(res2/b2)
     if (res2/b2 < eps2) then ! converged
        nitr = ncg; istat = 0
        exit cgloop
     end if

     call naxsl123pl(am,d,r,ru,ista2,iend1,jsta2,jend1,ksta2,kend1)

     call vdotv123pl(r,ru,rur1,ista2,iend1,jsta2,jend1,ksta2,kend1)

     beta = rur1/rur0
     rur0 = rur1
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              p(i,j,k) = ru(i,j,k)+beta*p(i,j,k)
           end do
        end do
     end do
! ==========================
!     iterative loop end
! ==========================
  end do cgloop

  return
end subroutine iccg123pl
! =============================================================================
subroutine ssorcg123pl(a,b,x,eps,nitr,istat,iprec &
     ,a4i,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     conjugate gradient method for SSOR-CG
!     solve 3-D Poisson equation by FDM
!
!     ver.0 on 20020823
!     ver.1 on 20020827; MPI parallelization
!     ver.2 on 20020828; some minor corrections for compatibilities
!                        with other linear solvers
!     ver.3 on 20030122
!     ver.4 on 20030206; domain partitioning by all three dimensions
!     ver.5 on 20030508; allocatable local arrays
!     ver.6 on 20030625; inline expansion of naxsub,vdotv,vnorm
!
!     global variables:
!       mx,my,mz: number of solution points in each direction
!       a(7,ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1):
!           nonzero elements of coefficient matrix
!       b(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1): RHS vector
!       x(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1):
!            initial guess on input, solution on output
!       a4i(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1):
!            inverse of diagonal elements of coefficient matrix
!       eps: tolerance on input, truncation error on output
!       nitr: maximum iteration on input, iteration number on output
!       r(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1):
!       p(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1):
!       ru(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1):
!       ap(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1):
!*********************************************************************

  integer :: i,j,k,ij
  integer :: ncg,maxitr,istat,iprec,nitr
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) &
       :: b,x,a4i,r,p,ru,ap
  real(dp) :: alpha,b2,beta,eps,eps2,rur0,rur1,pap,res2
  real(dp),allocatable :: &
       buffsi(:,:),buffri(:,:),buffsj(:,:),buffrj(:,:),buffsk(:,:),buffrk(:,:)

  maxitr = nitr; eps2 = eps**2

  allocate &
       (buffsi((jend1-jsta2+1)*(kend1-ksta2+1),2) &
       ,buffri((jend1-jsta2+1)*(kend1-ksta2+1),2) &
       ,buffsj((iend1-ista2+1)*(kend1-ksta2+1),2) &
       ,buffrj((iend1-ista2+1)*(kend1-ksta2+1),2) &
       ,buffsk((iend1-ista2+1)*(jend1-jsta2+1),2) &
       ,buffrk((iend1-ista2+1)*(jend1-jsta2+1),2))

! =======================
!     preconditioning
! =======================
  if (iprec == 0) then
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              a4i(i,j,k) = 1.e0_dp/a(i,j,k,4)
           enddo
        enddo
     enddo
     iprec = 1
  endif

! ----------------------
!     initialization
! ----------------------
  p = 0.e0_dp; r = 0.e0_dp; ru = 0.e0_dp; ap = 0.e0_dp

  call naxsub123pl(a,x,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k) = b(i,j,k)-ap(i,j,k)
        enddo
     enddo
  enddo

  call ssorsl123pl(a,a4i,r,ru &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
       ,buffri,buffsi,buffrj,buffsj,buffrk,buffsk)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           p(i,j,k) = ru(i,j,k)
        enddo
     enddo
  enddo
  call vdotv123pl(r,ru,rur0,ista2,iend1,jsta2,jend1,ksta2,kend1)

  call vnorm123pl(b,b2,ista2,iend1,jsta2,jend1,ksta2,kend1)

! ============================
!     iterative loop start
! ============================
  istat = 1
  cgloop: do ncg = 1,maxitr
!cdir iexpand
     call naxsub123pl(a,p,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)
!cdir iexpand
     call vdotv123pl(p,ap,pap,ista2,iend1,jsta2,jend1,ksta2,kend1)

     alpha = rur0/pap
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              x(i,j,k) = x(i,j,k)+alpha*p(i,j,k)
              r(i,j,k) = r(i,j,k)-alpha*ap(i,j,k)
           enddo
        enddo
     enddo
!cdir iexpand
     call vnorm123pl(r,res2,ista2,iend1,jsta2,jend1,ksta2,kend1)

     eps = dsqrt(res2/b2)
     if (res2/b2 <= eps2) then ! converged
        nitr = ncg; istat = 0
        exit cgloop
     endif

     call ssorsl123pl(a,a4i,r,ru &
          ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
          ,buffri,buffsi,buffrj,buffsj,buffrk,buffsk)
!cdir iexpand
     call vdotv123pl(r,ru,rur1,ista2,iend1,jsta2,jend1,ksta2,kend1)

     beta = rur1/rur0
     rur0 = rur1

     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              p(i,j,k) = ru(i,j,k)+beta*p(i,j,k)
           enddo
        enddo
     enddo

! ==========================
!     iterative loop end
! ==========================
  enddo cgloop

  deallocate(buffsi,buffri,buffsj,buffrj,buffsk,buffrk)

  return
end subroutine ssorcg123pl
! =============================================================================
subroutine pjcg123pl(a,b,x,eps,nitr,istat,iprec &
     ,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     conjugate gradient method with Point Jacobi preconditioning
!     solve 3-D Poisson equation by FDM
!
!     ver.0 on 20030604
!     ver.1 on 20030625: inline expansion of naxsub,vdotv,vnorm
!*********************************************************************

  integer :: i,j,k,ncg,maxitr,ij,nitr
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp) :: a(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7)
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: b,x,r,p,ru,ap,d
  real(dp) :: alpha,b2,beta,eps,eps2
  real(dp) :: rur0,rur1,pap,res2
  integer :: istat,iprec

  maxitr = nitr; eps2 = eps**2

! ===========================================
!     preconditioning by diagonal scaling
! ===========================================
  if (iprec == 0) then
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              d(i,j,k) = 1.e0_dp/a(i,j,k,4)
           end do
        end do
     end do
     iprec = 1
  end if

! ======================
!     initialization
! ======================
  p = 0.e0_dp; r = 0.e0_dp; ru = 0.e0_dp; ap = 0.e0_dp

  call naxsub123pl(a,x,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           r(i,j,k) = b(i,j,k)-ap(i,j,k)
        end do
     end do
  end do

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           ru(i,j,k) = d(i,j,k)*r(i,j,k)
        end do
     end do
  end do

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           p(i,j,k) = ru(i,j,k)
        end do
     end do
  end do
  call vdotv123pl(r,ru,rur0,ista2,iend1,jsta2,jend1,ksta2,kend1)

  call vnorm123pl(b,b2,ista2,iend1,jsta2,jend1,ksta2,kend1)

! ============================
!     iterative loop start
! ============================
  cgloop: do ncg = 1,maxitr
!cdir iexpand
     call naxsub123pl(a,p,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)
!cdir iexpand
     call vdotv123pl(p,ap,pap,ista2,iend1,jsta2,jend1,ksta2,kend1)

     alpha = rur0/pap
     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              x(i,j,k) = x(i,j,k)+alpha*p(i,j,k)
              r(i,j,k) = r(i,j,k)-alpha*ap(i,j,k)
           end do
        end do
     end do
!cdir iexpand
     call vnorm123pl(r,res2,ista2,iend1,jsta2,jend1,ksta2,kend1)

     eps = dsqrt(res2/b2)

     if (res2/b2.le.eps2) then ! converged
        nitr = ncg; istat = 0
        exit cgloop
     end if

     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              ru(i,j,k) = d(i,j,k)*r(i,j,k)
           end do
        end do
     end do
!cdir iexpand
     call vdotv123pl(r,ru,rur1,ista2,iend1,jsta2,jend1,ksta2,kend1)

     beta = rur1/rur0
     rur0 = rur1

     do k = ksta2,kend1
        do j = jsta2,jend1
           do i = ista2,iend1
              p(i,j,k) = ru(i,j,k)+beta*p(i,j,k)
           end do
        end do
     end do

! ==========================
!     iterative loop end
! ==========================
  end do cgloop

  return
end subroutine pjcg123pl
! =============================================================================
subroutine jacobi123p(a,b,x,eitr,nitr,istat &
     ,ista2,iend1,jsta2,jend1,ksta2,kend1)
  !*********************************************************************
  !     solve linear simultaneous equations (in 3-D FDM) by Jacobi method
  !
  !     domain partitioning in 3-directions and array reduction are performed
  !     ver.0 on 20030610
  !     ver.1 on 20030614: array size changed
  !     variables:
  !       a: non-zero elements of coefficients
  !       b: elements of right-hand side vector
  !       x: initial guess on input, solution on output
  !       istat: exit status; 0 for normal exit, 1 for abnormal exit
  !       nitr: maximun iteration on input, actual iteration on output
  !       eitr: maximum torrelance on input, residual on output
  !*********************************************************************

  integer :: i,j,k
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: &
       b,x,xold,a4i

  integer :: maxitr,nitr,istat
  real(dp) :: eitr,eps,zansa2,b2,zansa,zansa2tmp
  real(dp) :: z2tmp(0:nprocs-1)

  ! maximum torrelance
  eps = eitr; maxitr = nitr

  ! inverse of diagonal elements
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           a4i(i,j,k) = 1.e0_dp/a(i,j,k,4)
        enddo
     enddo
  enddo

  ! norm of right-hand side vector
  call vnorm123pl(b,b2,ista2,iend1,jsta2,jend1,ksta2,kend1)

! ==================
!     loop start
! ==================

  istat = 1
  itrloop: do nitr = 1,maxitr
     zansa2tmp = 0.e0_dp
     ! save guess at previous iteration
     xold = x
     call myshift123l(xold,ista2,iend1,jsta2,jend1,ksta2,kend1)

     do k = ksta2,kend1
        do j = jsta2,jend1
!cdir nodep
           do i = ista2,iend1
              zansa = b(i,j,k)-a(i,j,k,1)*xold(i,j,k-1) &
                   -a(i,j,k,2)*xold(i,j-1,k)-a(i,j,k,3)*xold(i-1,j,k) &
                   -a(i,j,k,4)*xold(i,j,k)-a(i,j,k,5)*xold(i+1,j,k) &
                   -a(i,j,k,6)*xold(i,j+1,k)-a(i,j,k,7)*xold(i,j,k+1)
              x(i,j,k) = xold(i,j,k)+zansa*a4i(i,j,k)
              zansa2tmp = zansa2tmp+zansa**2
           enddo
        enddo
     enddo

     call mpi_allgather(zansa2tmp,1,mpi_double_precision, &
          z2tmp,1,mpi_double_precision,mpi_comm_world,ierr)
     zansa2 = sum(z2tmp)
     eitr = dsqrt(zansa2/b2)

!     if (myrank == 0)write(*,'(i5,3e11.4)')nitr,eitr,zansa2,b2
     if (eitr < eps) then
        istat = 0; exit itrloop
     endif
  enddo itrloop
!
! ================
!     loop end
! ================
!
  return
end subroutine jacobi123p
! =============================================================================
subroutine sor2c123p(a,b,x,eitr,nitr,istat &
     ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
     ,arelax,buffri,buffsi,buffrj,buffsj,buffrk,buffsk)
!*********************************************************************
!     solve linear simultaneous equations (in 3-D FDM)
!     by Successive Over-Relaxation method with 2-color ordering
!     ver.0 on 20020808
!     ver.1 on 20020816
!     ver.2 on 20020819
!     ver.3 on 20020828
!     ver.4 on 20021129
!     ver.5 on 20030205; domain partitioning in 3-directions
!     ver.6 on 20030408; array reduction
!     ver.7 on 20030616; array size changed
!
!     variables:
!       a: non-zero elements of coefficients
!       b: elements of right-hand side vector
!       x: initial guess on input, solution on output
!       istat: exit status; 0 for normal exit, 1 for abnormal exit
!       nitr: maximun iteration on input, actual iteration on output
!       eitr: maximum torrelance on input, residual on output
!*********************************************************************

  integer :: i,j,k
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7), &
       intent(in) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: b,x

  integer :: maxsor,nsor,istat,nitr
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: &
       a4i,arelax
  real(dp) :: eitr,eps,zansa2,b2,zansa,zansa2tmp
  real(dp),dimension((jend1-jsta2+1)*(kend1-ksta2+1),2) :: buffri,buffsi
  real(dp),dimension((iend1-ista2+1)*(kend1-ksta2+1),2) :: buffrj,buffsj
  real(dp),dimension((iend1-ista2+1)*(jend1-jsta2+1),2) :: buffrk,buffsk
  real(dp) :: z2tmp(0:nprocs-1)

  ! maximum torrelance
  eps = eitr; maxsor = nitr

  ! inverse of diagonal elements
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           a4i(i,j,k) = 1.e0_dp/a(i,j,k,4)
        enddo
     enddo
  enddo

  ! norm of right-hand side vector
  call vnorm123pl(b,b2,ista2,iend1,jsta2,jend1,ksta2,kend1)

! ==================
!     loop start
! ==================

  istat = 1; nitr = maxsor
  sorloop: do nsor = 1,maxsor
     zansa2tmp = 0.e0_dp
     call shift4sor2c123pl(x,0, &
          ista2,iend1,jsta2,jend1,ksta2,kend1, &
          buffri,buffsi,buffrj,buffsj,buffrk,buffsk)

     ! 1st color (for even i+j+k)
     do k = ksta2,kend1
        do j = jsta2,jend1
!cdir nodep
           do i = ista2+mod(ista2+j+k,2),iend1,2
              zansa = b(i,j,k)-a(i,j,k,1)*x(i,j,k-1) &
                   -a(i,j,k,2)*x(i,j-1,k)-a(i,j,k,3)*x(i-1,j,k) &
                   -a(i,j,k,4)*x(i,j,k)-a(i,j,k,5)*x(i+1,j,k) &
                   -a(i,j,k,6)*x(i,j+1,k)-a(i,j,k,7)*x(i,j,k+1)
              x(i,j,k) = x(i,j,k)+arelax(i,j,k)*zansa*a4i(i,j,k)
              zansa2tmp = zansa2tmp+zansa**2
           enddo
        enddo
     enddo

     call shift4sor2c123pl(x,1, &
          ista2,iend1,jsta2,jend1,ksta2,kend1, &
          buffri,buffsi,buffrj,buffsj,buffrk,buffsk)

     ! 2nd color (for odd i+j+k)
     do k = ksta2,kend1
        do j = jsta2,jend1
!cdir nodep
           do i = ista2+mod(ista2+j+k+1,2),iend1,2
              zansa = b(i,j,k)-a(i,j,k,1)*x(i,j,k-1) &
                   -a(i,j,k,2)*x(i,j-1,k)-a(i,j,k,3)*x(i-1,j,k) &
                   -a(i,j,k,4)*x(i,j,k)-a(i,j,k,5)*x(i+1,j,k) &
                   -a(i,j,k,6)*x(i,j+1,k)-a(i,j,k,7)*x(i,j,k+1)
              x(i,j,k) = x(i,j,k)+arelax(i,j,k)*zansa*a4i(i,j,k)
              zansa2tmp = zansa2tmp+zansa**2
           enddo
        enddo
     enddo

     call mpi_allgather(zansa2tmp,1,mpi_double_precision, &
          z2tmp,1,mpi_double_precision,mpi_comm_world,ierr)
     zansa2 = sum(z2tmp)
     eitr = dsqrt(zansa2/b2)

!     if (myrank == 0)write(*,'(i5,3e11.4)')nsor,eitr,zansa2,b2
     if (eitr < eps) then
        istat = 0; nitr = nsor; exit sorloop
     endif
  enddo sorloop

! ================
!     loop end
! ================

  return
end subroutine sor2c123p
! =============================================================================
subroutine liludcmp123pl(a,am,d,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     define/decompose preconditioner for localized ILU(0)
!*********************************************************************

  integer :: i,j,k
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a,am
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: d
  real(dp) :: w

! -------------------------
!     initialize arrays
! -------------------------
  am = 0.e0_dp; d = 0.e0_dp

! ====================================
!     define preconditioner matrix
! ====================================

  am = a

!     zeroing out
  do j = jsta2,jend1
     do i = ista2,iend1
        am(i,j,ksta2,1) = 0.e0_dp; am(i,j,kend1,7) = 0.e0_dp
     enddo
  enddo
!
  do k = ksta2,kend1
     do i = ista2,iend1
        am(i,jsta2,k,2) = 0.e0_dp; am(i,jend1,k,6) = 0.e0_dp
     enddo
  enddo
!
  do k = ksta2,kend1
     do j = jsta2,jend1
        am(ista2,j,k,3) = 0.e0_dp; am(iend1,j,k,5) = 0.e0_dp
     enddo
  enddo

! =======================================
!     decompose preconditioner matrix
! =======================================

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           w = am(i,j,k,4) - am(i,j,k,3)*am(i-1,j,k,5)*d(i-1,j,k) &
                - am(i,j,k,2)*am(i,j-1,k,6)*d(i,j-1,k) &
                - am(i,j,k,1)*am(i,j,k-1,7)*d(i,j,k-1)
           d(i,j,k) = 1.e0_dp/w
        enddo
     enddo
  enddo

  return
end subroutine liludcmp123pl
! =============================================================================
subroutine myshift4a123l(a,ista2,iend1,jsta2,jend1,ksta2,kend1)

  integer,intent(in) ::  ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp) :: a(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7)

  call myshift123l(a(ista2-1,jsta2-1,ksta2-1,1), &
       ista2,iend1,jsta2,jend1,ksta2,kend1)
  call myshift123l(a(ista2-1,jsta2-1,ksta2-1,2), &
       ista2,iend1,jsta2,jend1,ksta2,kend1)
  call myshift123l(a(ista2-1,jsta2-1,ksta2-1,3), &
       ista2,iend1,jsta2,jend1,ksta2,kend1)
  call myshift123l(a(ista2-1,jsta2-1,ksta2-1,4), &
       ista2,iend1,jsta2,jend1,ksta2,kend1)
  call myshift123l(a(ista2-1,jsta2-1,ksta2-1,5), &
       ista2,iend1,jsta2,jend1,ksta2,kend1)
  call myshift123l(a(ista2-1,jsta2-1,ksta2-1,6), &
       ista2,iend1,jsta2,jend1,ksta2,kend1)
  call myshift123l(a(ista2-1,jsta2-1,ksta2-1,7), &
       ista2,iend1,jsta2,jend1,ksta2,kend1)

  return
end subroutine myshift4a123l
! =============================================================================
subroutine shift4sor2c123pl(p,iflag &
     ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
     ,buffri,buffsi,buffrj,buffsj,buffrk,buffsk)
!*********************************************************************
!     shift data for Red-Black SOR
!
!     iflg: = 0 when called before calculating 1st color
!           = 1 when called before calculating 2nd color
!*********************************************************************

  integer :: i,j,k,ij
  integer,intent(in) :: iflag
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: p
  integer :: ireqsi1,ireqsi2,ireqri1,ireqri2
  integer :: ireqsj1,ireqsj2,ireqrj1,ireqrj2
  integer :: ireqsk1,ireqsk2,ireqrk1,ireqrk2
  integer :: iup,idown,jup,jdown,kup,kdown
  integer :: icnti1,icnti2,icntj1,icntj2,icntk1,icntk2
  real(dp),dimension((jend1-jsta2+1)*(kend1-ksta2+1),2) :: buffri,buffsi
  real(dp),dimension((iend1-ista2+1)*(kend1-ksta2+1),2) :: buffrj,buffsj
  real(dp),dimension((iend1-ista2+1)*(jend1-jsta2+1),2) :: buffrk,buffsk
  integer :: nbufi,nbufj,nbufk

  iup = itable(myranki+1,myrankj,myrankk)
  idown = itable(myranki-1,myrankj,myrankk)
  jup = itable(myranki,myrankj+1,myrankk)
  jdown = itable(myranki,myrankj-1,myrankk)
  kup = itable(myranki,myrankj,myrankk+1)
  kdown = itable(myranki,myrankj,myrankk-1)
  nbufi = (jend1-jsta2+1)*(kend1-ksta2+1)
  nbufj = (iend1-ista2+1)*(kend1-ksta2+1)
  nbufk = (iend1-ista2+1)*(jend1-jsta2+1)

  if (myranki /= 0) then
     icnti1 = 0
     do k = ksta2,kend1
        do j = jsta2+mod(ista2+jsta2+k+1+iflag,2),jend1,2
           icnti1 = icnti1+1
           buffsi(icnti1,1) = p(ista2,j,k)
        enddo
     enddo
     call mpi_send(buffsi(1,1),icnti1,mpi_double_precision,idown,1, &
          mpi_comm_world,ierr)
  endif
  if (myranki /= iprocs-1) then
     call mpi_recv(buffri(1,2),nbufi, mpi_double_precision,iup,  1, &
          mpi_comm_world,istatus,ierr)
     ij = 0
     do k = ksta2,kend1
        do j = jsta2+mod(iend1+jsta2+k+iflag,2),jend1,2
           ij = ij+1
           p(iend1+1,j,k) = buffri(ij,2)
        enddo
     enddo
  endif

  if (myranki /= iprocs-1) then
     icnti2 = 0
     do k = ksta2,kend1
        do j = jsta2+mod(iend1+jsta2+k+1+iflag,2),jend1,2
           icnti2 = icnti2+1
           buffsi(icnti2,2) = p(iend1,j,k)
        enddo
     enddo
     call mpi_send(buffsi(1,2),icnti2,mpi_double_precision,iup,  1, &
          mpi_comm_world,ierr)
  endif
  if (myranki /= 0) then
     call mpi_recv(buffri(1,1),nbufi, mpi_double_precision,idown,1, &
          mpi_comm_world,istatus,ierr)
     ij = 0
     do k = ksta2,kend1
        do j = jsta2+mod(ista2+jsta2+k+iflag,2),jend1,2
           ij = ij+1
           p(ista2-1,j,k) = buffri(ij,1)
        enddo
     enddo
  endif

  if (myrankj /= 0) then
     icntj1 = 0
     do k = ksta2,kend1
        do i = ista2+mod(ista2+jsta2+k+1+iflag,2),iend1,2
           icntj1 = icntj1+1
           buffsj(icntj1,1) = p(i,jsta2,k)
        enddo
     enddo
     call mpi_send(buffsj(1,1),icntj1,mpi_double_precision,jdown,1, &
          mpi_comm_world,ierr)
  endif
  if (myrankj /= jprocs-1) then
     call mpi_recv(buffrj(1,2),nbufj, mpi_double_precision,jup,  1, &
          mpi_comm_world,istatus,ierr)
     ij = 0
     do k = ksta2,kend1
        do i = ista2+mod(ista2+jend1+k+iflag,2),iend1,2
           ij = ij+1
           p(i,jend1+1,k) = buffrj(ij,2)
        enddo
     enddo
  endif

  if (myrankj /= jprocs-1) then
     icntj2 = 0
     do k = ksta2,kend1
        do i = ista2+mod(ista2+jend1+k+1+iflag,2),iend1,2
           icntj2 = icntj2+1
           buffsj(icntj2,2) = p(i,jend1,k)
        enddo
     enddo
     call mpi_send(buffsj(1,2),icntj2,mpi_double_precision,jup,  1, &
          mpi_comm_world,ierr)
  endif
  if (myrankj /= 0) then
     call mpi_recv(buffrj(1,1),nbufj, mpi_double_precision,jdown,1, &
          mpi_comm_world,istatus,ierr)
     ij = 0
     do k = ksta2,kend1
        do i = ista2+mod(ista2+jsta2+k+iflag,2),iend1,2
           ij = ij+1
           p(i,jsta2-1,k) = buffrj(ij,1)
        enddo
     enddo
  endif

  if (myrankk /= 0) then
     icntk1 = 0
     do j = jsta2,jend1
        do i = ista2+mod(ista2+j+ksta2+1+iflag,2),iend1,2
           icntk1 = icntk1+1
           buffsk(icntk1,1) = p(i,j,ksta2)
        enddo
     enddo
     call mpi_send(buffsk(1,1),icntk1,mpi_double_precision,kdown,1, &
          mpi_comm_world,ierr)
  endif
  if (myrankk /= kprocs-1) then
     call mpi_recv(buffrk(1,2),nbufk, mpi_double_precision,kup,  1, &
          mpi_comm_world,istatus,ierr)
     ij = 0
     do j = jsta2,jend1
        do i = ista2+mod(ista2+j+kend1+iflag,2),iend1,2
           ij = ij+1
           p(i,j,kend1+1) = buffrk(ij,2)
        enddo
     enddo
  endif

  if (myrankk /= kprocs-1) then
     icntk2 = 0
     do j = jsta2,jend1
        do i = ista2+mod(ista2+j+kend1+1+iflag,2),iend1,2
           icntk2 = icntk2+1
           buffsk(icntk2,2) = p(i,j,kend1)
        enddo
     enddo
     call mpi_send(buffsk(1,2),icntk2,mpi_double_precision,kup,  1, &
          mpi_comm_world,ierr)
  endif
  if (myrankk /= 0) then
     call mpi_recv(buffrk(1,1),nbufk, mpi_double_precision,kdown,1, &
          mpi_comm_world,istatus,ierr)
     ij = 0
     do j = jsta2,jend1
        do i = ista2+mod(ista2+j+ksta2+iflag,2),iend1,2
           ij = ij+1
           p(i,j,ksta2-1) = buffrk(ij,1)
        enddo
     enddo
  endif

 return
end subroutine shift4sor2c123pl
! =============================================================================
subroutine ssorsl123pl(a,a4i,b,x &
     ,ista2,iend1,jsta2,jend1,ksta2,kend1 &
     ,buffri,buffsi,buffrj,buffsj,buffrk,buffsk)
!*********************************************************************
!     preconditioner for CG solving 3-D Poisson equation by FDM
!     Symmetric Successive Over Relaxation method by 2 coloring
!
!     ver.0 on 20020823; as stand-alone SSOR solver for Ax=b
!     ver.1 on 20030206; domain partitioning by all three dimensions
!     ver.2 on 20030616; array size changed
!
!     global variables:
!       a: non-zero elements of coefficients
!       b: elements of right-hand side vector
!       x: initial guess on input, solution on output
!       mx,my,mz: number of solution points in each direction
!     local variables
!       maxsor: maximun iteration number
!       eps: maximum tolerance
!       relc: relaxation coefficient
!       eitr,eitrtmp: residual
!*********************************************************************

  integer :: i,j,k,ij
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: b,x,a4i
  real(dp),dimension((jend1-jsta2+1)*(kend1-ksta2+1),2) :: buffri,buffsi
  real(dp),dimension((iend1-ista2+1)*(kend1-ksta2+1),2) :: buffrj,buffsj
  real(dp),dimension((iend1-ista2+1)*(jend1-jsta2+1),2) :: buffrk,buffsk

! initial guess of solution
  x = 0.e0_dp

! +++++++++++++++++++
!     forward SOR
! +++++++++++++++++++
!
!     1st color (for even i+j+k)
!
!cdir nosync
  do k = ksta2,kend1
     do j = jsta2,jend1
!cdir nodep
        do i = ista2+mod(ista2+j+k,2),iend1,2
           x(i,j,k) = (b(i,j,k) &
                -a(i,j,k,1)*x(i,j,k-1)-a(i,j,k,2)*x(i,j-1,k) &
                -a(i,j,k,3)*x(i-1,j,k)-a(i,j,k,5)*x(i+1,j,k) &
                -a(i,j,k,6)*x(i,j+1,k)-a(i,j,k,7)*x(i,j,k+1) &
                )*a4i(i,j,k)
        enddo
     enddo
  enddo
!
!     2nd color (for odd i+j+k)
!
  call shift4sor2c123pl(x,1, &
       ista2,iend1,jsta2,jend1,ksta2,kend1,&
       buffri,buffsi,buffrj,buffsj,buffrk,buffsk)
!cdir nosync
  do k = ksta2,kend1
     do j = jsta2,jend1
!cdir nodep
        do i = ista2+mod(ista2+j+k+1,2),iend1,2
           x(i,j,k) = (b(i,j,k) &
                -a(i,j,k,1)*x(i,j,k-1)-a(i,j,k,2)*x(i,j-1,k) &
                -a(i,j,k,3)*x(i-1,j,k)-a(i,j,k,5)*x(i+1,j,k) &
                -a(i,j,k,6)*x(i,j+1,k)-a(i,j,k,7)*x(i,j,k+1) &
                )*a4i(i,j,k)
        enddo
     enddo
  enddo
!
! ++++++++++++++++++++
!     backward SOR
! ++++++++++++++++++++
!
!     2nd color (for odd i+j+k)
!
!cdir nosync
  do k = kend1,ksta2,-1
     do j = jend1,jsta2,-1
!cdir nodep
        do i = iend1-mod(iend1+j+k+1,2),ista2,-2
           x(i,j,k) = (b(i,j,k) &
                -a(i,j,k,1)*x(i,j,k-1)-a(i,j,k,2)*x(i,j-1,k) &
                -a(i,j,k,3)*x(i-1,j,k)-a(i,j,k,5)*x(i+1,j,k) &
                -a(i,j,k,6)*x(i,j+1,k)-a(i,j,k,7)*x(i,j,k+1) &
                )*a4i(i,j,k)
        enddo
     enddo
  enddo
!
!     1st color (for even i+j+k)
!
  call shift4sor2c123pl(x,0 &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1, &
       buffri,buffsi,buffrj,buffsj,buffrk,buffsk)
!cdir nosync
  do k = kend1,ksta2,-1
     do j = jend1,jsta2,-1
!cdir nodep
        do i = iend1-mod(iend1+j+k,2),ista2,-2
           x(i,j,k) = (b(i,j,k) &
                -a(i,j,k,1)*x(i,j,k-1)-a(i,j,k,2)*x(i,j-1,k) &
                -a(i,j,k,3)*x(i-1,j,k)-a(i,j,k,5)*x(i+1,j,k) &
                -a(i,j,k,6)*x(i,j+1,k)-a(i,j,k,7)*x(i,j,k+1) &
                )*a4i(i,j,k)
        enddo
     enddo
  enddo

  return
end subroutine ssorsl123pl
! =============================================================================
subroutine natsl123pl(am,d,b,x,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     Solve (LU)t*x=b
!     where (LU)t is transpose of matrix (LU)
!     vector b is RHS vector on input, solution on output
!*********************************************************************

  integer :: i,j,k
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: am
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: d,b,x

  x = 0.e0_dp

  ! forward substitution
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           x(i,j,k) = b(i,j,k) &
                -am(i-1,j,k,5)*d(i-1,j,k)*x(i-1,j,k) &
                -am(i,j-1,k,6)*d(i,j-1,k)*x(i,j-1,k) &
                -am(i,j,k-1,7)*d(i,j,k-1)*x(i,j,k-1)
        enddo
     enddo
  enddo

!     backward substitution
  do k = kend1,ksta2,-1
     do j = jend1,jsta2,-1
        do i = iend1,ista2,-1
           x(i,j,k) = d(i,j,k)*x(i,j,k) &
                -d(i,j,k)*am(i+1,j,k,3)*x(i+1,j,k) &
                -d(i,j,k)*am(i,j+1,k,2)*x(i,j+1,k) &
                -d(i,j,k)*am(i,j,k+1,1)*x(i,j,k+1)
        enddo
     enddo
  enddo

  return
end subroutine natsl123pl
! =============================================================================
subroutine natsub123pl(a,p,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     Compute product
!     (vector AP) = (matrix At)*(vector  P)
!     where (At) is transpose of (matrix A)
!*********************************************************************

  integer :: i,j,k,ij
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: p,ap

  call myshift123l(p,ista2,iend1,jsta2,jend1,ksta2,kend1)

  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           ap(i,j,k) = a(i,j,k-1,7)*p(i,j,k-1) &
                +a(i,j-1,k,6)*p(i,j-1,k) &
                +a(i-1,j,k,5)*p(i-1,j,k) &
                +a(i,j,k,4)*p(i,j,k) &
                +a(i+1,j,k,3)*p(i+1,j,k) &
                +a(i,j+1,k,2)*p(i,j+1,k) &
                +a(i,j,k+1,1)*p(i,j,k+1)
        enddo
     enddo
  enddo

  return
end subroutine natsub123pl
! =============================================================================
subroutine naxsl123pl(am,d,b,x,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     Solve (LU)*x=b
!     vector b is RHS vector on input, solution on output
!     Localized ILU(0) preconditioning
!*********************************************************************

  integer :: i,j,k
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: am
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: d,b,x

  x = 0.e0_dp

  ! forward substitution
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           ! x(i,j,k) = d(i,j,k)*(b(i,j,k)-am(i,j,k,3)*x(i-1,j,k) &
           !     -am(i,j,k,2)*x(i,j-1,k)-am(i,j,k,1)*x(i,j,k-1))
           x(i,j,k) = d(i,j,k)*b(i,j,k) &
                -d(i,j,k)*am(i,j,k,3)*x(i-1,j,k) &
                -d(i,j,k)*am(i,j,k,2)*x(i,j-1,k) &
                -d(i,j,k)*am(i,j,k,1)*x(i,j,k-1)
        enddo
     enddo
  enddo

  ! backward substitution
  do k = kend1,ksta2,-1
     do j = jend1,jsta2,-1
        do i = iend1,ista2,-1
           ! x(i,j,k) = x(i,j,k)-d(i,j,k)*(am(i,j,k,5)*x(i+1,j,k)&
           !     +am(i,j,k,6)*x(i,j+1,k)+am(i,j,k,7)*x(i,j,k+1))
           x(i,j,k) = x(i,j,k) &
                -d(i,j,k)*am(i,j,k,5)*x(i+1,j,k) &
                -d(i,j,k)*am(i,j,k,6)*x(i,j+1,k) &
                -d(i,j,k)*am(i,j,k,7)*x(i,j,k+1)
        enddo
     enddo
  enddo

  return
end subroutine naxsl123pl
! =============================================================================
subroutine naxsub123pl(a,p,ap,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     Compute product
!     (vector AP) = (matrix A)*(vector  P)
!*********************************************************************

  integer :: i,j,k
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1,7) :: a
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: p,ap
  integer :: jtmp

  call myshift123l(p,ista2,iend1,jsta2,jend1,ksta2,kend1)
  jtmp = jend1-mod(jend1-jsta2+1,2) ! 2-step unrolling in j's loop
  do k = ksta2,kend1
     do j = jsta2,jtmp,2
        do i = ista2,iend1
           ap(i,j,k) = a(i,j,k,1)*p(i,j,k-1)+a(i,j,k,2)*p(i,j-1,k) &
                +a(i,j,k,3)*p(i-1,j,k)+a(i,j,k,4)*p(i,j,k) &
                +a(i,j,k,5)*p(i+1,j,k)+a(i,j,k,6)*p(i,j+1,k) &
                +a(i,j,k,7)*p(i,j,k+1)
           ap(i,j+1,k) = a(i,j+1,k,1)*p(i,j+1,k-1)+a(i,j+1,k,2)*p(i,j,k) &
                +a(i,j+1,k,3)*p(i-1,j+1,k)+a(i,j+1,k,4)*p(i,j+1,k) &
                +a(i,j+1,k,5)*p(i+1,j+1,k)+a(i,j+1,k,6)*p(i,j+2,k) &
                +a(i,j+1,k,7)*p(i,j+1,k+1)
        enddo
     enddo
     do j = jtmp+1,jend1
        do i = ista2,iend1
           ap(i,j,k) = a(i,j,k,1)*p(i,j,k-1)+a(i,j,k,2)*p(i,j-1,k) &
                +a(i,j,k,3)*p(i-1,j,k)+a(i,j,k,4)*p(i,j,k) &
                +a(i,j,k,5)*p(i+1,j,k)+a(i,j,k,6)*p(i,j+1,k) &
                +a(i,j,k,7)*p(i,j,k+1)
        enddo
     enddo
  enddo

  return
end subroutine naxsub123pl
! =============================================================================
subroutine vdotv123pl(x,y,r,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     Compute inner product of vectors x and y
!*********************************************************************

  integer :: i,j,k
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: x,y
  real(dp),intent(out) :: r
  real(dp) :: rtmp1,rtmp2(0:nprocs-1)

  rtmp1 = sum(x(ista2:iend1,jsta2:jend1,ksta2:kend1) &
       *y(ista2:iend1,jsta2:jend1,ksta2:kend1))
  call mpi_allgather(rtmp1,1,mpi_double_precision, &
       rtmp2,1,mpi_double_precision,mpi_comm_world,ierr)
  r = sum(rtmp2)

  return
end subroutine vdotv123pl
! =============================================================================
subroutine vnorm123pl(x,x2,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!     Compute square of norm of vector x
!*********************************************************************

  integer :: i,j,k
  integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
  real(dp),dimension(ista2-1:iend1+1,jsta2-1:jend1+1,ksta2-1:kend1+1) :: x
  real(dp),intent(out) :: x2
  real(dp) :: x2tmp1,x2tmp2(0:nprocs-1)

  x2tmp1 = sum(x(ista2:iend1,jsta2:jend1,ksta2:kend1)**2)
  call mpi_allgather(x2tmp1,1,mpi_double_precision, &
       x2tmp2,1,mpi_double_precision,mpi_comm_world,ierr)
  x2 = sum(x2tmp2)

  return
end subroutine vnorm123pl
! =============================================================================
end module mylinsol
