!option! -pvctl vecthreshold
!*********************************************************************
!     This file contains common tools for MPI calculations
!     Domain splitting by all three dimensions is assumed
!     ver.0 on 20030205
!     ver.1 on 20030605
!     ver.2 on 20030609
!     ver.3 on 20030613
!     ver.4 on 20030616
!     ver.5 on 20030618: module is renamed as mympitools
!*********************************************************************
module mympitools
  use constants
  implicit none
  include 'mpif.h'
  integer :: ierr,myrank,nprocs,istatus(mpi_status_size)
  integer :: myranki,myrankj,myrankk
  include "mpiprocnum.h"
  integer :: itable(-1:iprocs,-1:jprocs,-1:kprocs)
  integer,dimension(0:iprocs-1) :: istas,iends,inums
  integer,dimension(0:jprocs-1) :: jstas,jends,jnums
  integer,dimension(0:kprocs-1) :: kstas,kends,knums

contains
! --------------------------------------------------------------------
  subroutine mympiinit
    call mpi_init(ierr)
    call mpi_comm_size(mpi_comm_world,nprocs,ierr)
    call mpi_comm_rank(mpi_comm_world,myrank,ierr)
    return
  end subroutine mympiinit
! --------------------------------------------------------------------
  subroutine mympifinalize
    call mpi_finalize(ierr)
    return
  end subroutine mympifinalize
! --------------------------------------------------------------------
  subroutine para_range123(i1,i2,j1,j2,k1,k2)
    integer,intent(in) :: i1,i2,j1,j2,k1,k2
    integer :: irank,jrank,krank
    call para_range(i1,i2,istas,iends,inums,iprocs)
    call para_range(j1,j2,jstas,jends,jnums,jprocs)
    call para_range(k1,k2,kstas,kends,knums,kprocs)
    return
  end subroutine para_range123
! --------------------------------------------------------------------
  subroutine para_range(n1,n2,nstas,nends,nnums,ndivs)
!*********************************************************************
!     splits do-loop iteration range into NDIV's segments
!*********************************************************************
    integer,intent(in) :: ndivs,n1,n2
    integer,dimension(0:ndivs-1),intent(out) :: nstas,nends,nnums
    integer :: irank,iwork1,iwork2,iwork3,iwork4,nall
    nall = n2-n1+1
    iwork1 = nall/ndivs+1
    iwork2 = iwork1*ndivs-nall
    nnums = iwork1
    do irank = min(1,ndivs-iwork2),min(iwork2,ndivs-1)
       nnums(irank) = iwork1-1
    end do
    nstas(0) = n1
    do irank = 1,ndivs-1
       nstas(irank) = nstas(irank-1)+nnums(irank-1)
    end do
    do irank = 0,ndivs-1
       nends(irank) = nstas(irank)+nnums(irank)-1
    end do
    return
  end subroutine para_range
! --------------------------------------------------------------------
  subroutine proc_table
!*********************************************************************
!     generate process table
!*********************************************************************
    integer :: i,j,k,irank
    if (nprocs /= iprocs*jprocs*kprocs) then
       write(*,'(4a10)')"nprocs","iprocs","jprocs","kprocs"
       write(*,'(4i10)')nprocs,iprocs,jprocs,kprocs
       stop
    end if
    itable = mpi_proc_null
    do k = 0,kprocs-1
       do j = 0,jprocs-1
          do i = 0,iprocs-1
             irank = i+(j+k*jprocs)*iprocs
             itable(i,j,k) = irank
             if (myrank == irank) then
                myranki = i; myrankj = j; myrankk = k
             end if
          end do
       end do
    end do
    return
  end subroutine proc_table
! --------------------------------------------------------------------
  subroutine mygather123lg(xl,xg,ista,iend,jsta,jend,ksta,kend,iax,jax,kax)
!*********************************************************************
!     gather distributed data to process 0
!*********************************************************************
    integer :: i,j,k,ijk
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend,iax,jax,kax
    real(dp),intent(in) :: xl(iend-ista+1,jend-jsta+1,kend-ksta+1)
    real(dp),intent(out) :: xg(iax,jax,kax)
    integer,dimension(0:iprocs-1,0:jprocs-1,0:kprocs-1) :: ijlen,kdisp,iireq
    integer :: irank,jrank,krank,iorig,ireq
#ifdef ES
    real(dp),allocatable,save :: buff(:)
    integer,save :: alloc_buffer=0
#else
    real(dp),allocatable :: buff(:)
#endif
    integer :: inum,jnum,knum
    integer :: istad,iendd,jstad,jendd,kstad,kendd,inumd,jnumd,knumd
    inum = iend-ista+1; jnum = jend-jsta+1; knum = kend-ksta+1
    do krank = 0,kprocs-1
       kstad=kstas(krank); kendd=kends(krank); knumd=kendd-kstad+1
       do jrank = 0,jprocs-1
          jstad=jstas(jrank); jendd=jends(jrank); jnumd=jendd-jstad+1
          do irank = 0,iprocs-1
             istad=istas(irank); iendd=iends(irank); inumd=iendd-istad+1
             ijlen(irank,jrank,krank) = inumd*jnumd*knumd
             kdisp(irank,jrank,krank) = 1 &
                  +iax*jax*(kstad-1)+iax*(jstad-1)*knumd+(istad-1)*jnumd*knumd
          end do
       end do
    end do
    ! allocate buffer
#ifdef ES
    if (alloc_buffer == 0) then
       allocate(buff(iax*jax*kax))
       alloc_buffer = 1
    end if
#else
    allocate(buff(iax*jax*kax))
#endif
    ! set sending buffer
    do k = 1,knum
       do j = 1,jnum
          do i = 1,inum
             ijk = kdisp(myranki,myrankj,myrankk) &
                  +(i-1)+(j-1)*inum+(k-1)*inum*jnum
             buff(ijk) = xl(i,j,k)
          end do
       end do
    end do
    if (myrank == 0) then
       do krank = 0,kprocs-1
          do jrank = 0,jprocs-1
             do irank = 0,iprocs-1
                iorig = itable(irank,jrank,krank)
                if (iorig == 0) iorig = mpi_proc_null
                call mpi_irecv(buff(kdisp(irank,jrank,krank)) &
                     ,ijlen(irank,jrank,krank) &
                     ,mpi_double_precision,iorig,1 &
                     ,mpi_comm_world,iireq(irank,jrank,krank),ierr)
             end do
          end do
       end do
       do krank = 0,kprocs-1
          do jrank = 0,jprocs-1
             do irank = 0,iprocs-1
                call mpi_wait(iireq(irank,jrank,krank),istatus,ierr)
             end do
          end do
       end do
    else
       call mpi_isend(buff(kdisp(myranki,myrankj,myrankk)), &
            ijlen(myranki,myrankj,myrankk), &
            mpi_double_precision, &
            0,1,mpi_comm_world,ireq,ierr)
       call mpi_wait(ireq,istatus,ierr)
    end if
    if (myrank == 0) then
       do krank = 0,kprocs-1
          kstad=kstas(krank); kendd=kends(krank); knumd=kendd-kstad+1
          do jrank = 0,jprocs-1
             jstad=jstas(jrank); jendd=jends(jrank); jnumd=jendd-jstad+1
             do irank = 0,iprocs-1
                istad=istas(irank); iendd=iends(irank); inumd=iendd-istad+1
                do k = kstad,kendd
                   do j = jstad,jendd
                      do i = istad,iendd
                         ijk = kdisp(irank,jrank,krank) &
                              +(i-istad)+(j-jstad)*inumd+(k-kstad)*inumd*jnumd
                         xg(i,j,k) = buff(ijk)
                      end do
                   end do
                end do
             end do
          end do
       end do
    end if
#ifndef ES
    deallocate(buff)
#endif
    return
  end subroutine mygather123lg
! --------------------------------------------------------------------
  subroutine myooban123lg(xl,xg,ista,iend,jsta,jend,ksta,kend,iax,jax,kax)
!*********************************************************************
!*********************************************************************
    integer :: i,j,k,ijk
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend,iax,jax,kax
    real(dp),intent(in) :: xl(iend-ista+1,jend-jsta+1,kend-ksta+1)
    real(dp),intent(out) :: xg(iax,jax,kax)
    integer,dimension(0:iprocs-1,0:jprocs-1,0:kprocs-1) :: ijlen,kdisp
    integer :: irank,jrank,krank
#ifdef ES
    real(dp),allocatable,save :: buff(:)
    integer,save :: alloc_buffer=0
#else
    real(dp),allocatable :: buff(:)
#endif
    integer :: inum,jnum,knum
    integer :: istad,iendd,jstad,jendd,kstad,kendd,inumd,jnumd,knumd
    inum = iend-ista+1; jnum = jend-jsta+1; knum = kend-ksta+1
    do krank = 0,kprocs-1
       kstad=kstas(krank); kendd=kends(krank); knumd=kendd-kstad+1
       do jrank = 0,jprocs-1
          jstad=jstas(jrank); jendd=jends(jrank); jnumd=jendd-jstad+1
          do irank = 0,iprocs-1
             istad=istas(irank); iendd=iends(irank); inumd=iendd-istad+1
             ijlen(irank,jrank,krank) = inumd*jnumd*knumd
             kdisp(irank,jrank,krank) = 1 &
                  +iax*jax*(kstad-1)+iax*(jstad-1)*knumd+(istad-1)*jnumd*knumd
          end do
       end do
    end do
    ! allocate buffer
#ifdef ES
    if (alloc_buffer == 0) then
       allocate (buff(iax*jax*kax))
       alloc_buffer = 1
    end if
#else
    allocate (buff(iax*jax*kax))
#endif
    ! set sending buffer
    do k = 1,knum
       do j = 1,jnum
          do i = 1,inum
             ijk = kdisp(myranki,myrankj,myrankk) &
                  +(i-1)+(j-1)*inum+(k-1)*inum*jnum
             buff(ijk) = xl(i,j,k)
          end do
       end do
    end do
    do krank = 0,kprocs-1
       do jrank = 0,jprocs-1
          do irank = 0,iprocs-1
             call mpi_bcast(buff(kdisp(irank,jrank,krank)) &
                  ,ijlen(irank,jrank,krank), &
                  mpi_double_precision, &
                  itable(irank,jrank,krank),mpi_comm_world,ierr)
          end do
       end do
    end do
    do krank = 0,kprocs-1
       kstad = kstas(krank); kendd = kends(krank); knumd = kendd-kstad+1
       do jrank = 0,jprocs-1
          jstad = jstas(jrank); jendd = jends(jrank); jnumd = jendd-jstad+1
          do irank = 0,iprocs-1
             istad = istas(irank); iendd = iends(irank); inumd = iendd-istad+1
             do k = kstad,kendd
                do j = jstad,jendd
                   do i = istad,iendd
                      ijk = kdisp(irank,jrank,krank) &
                           +(i-istad)+(j-jstad)*inumd+(k-kstad)*inumd*jnumd
                      xg(i,j,k) = buff(ijk)
                   end do
                end do
             end do
          end do
       end do
    end do
#ifndef ES
    deallocate(buff)
#endif
    return
  end subroutine myooban123lg
! --------------------------------------------------------------------
  subroutine mypartialshift1l1(x,ista,iend,jsta,jend,ksta,kend)
!*********************************************************************
! send to iup | recv from idown
!*********************************************************************
    integer :: i,j,k
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: x(0:iend-ista+1+1,0:jend-jsta+1+1,0:kend-ksta+1+1)
    integer :: inum,jnum,knum,iup,idown,ireqs,ireqr,ndata
#ifdef ES
    real(dp),allocatable,save :: buffs(:),buffr(:)
    integer,save :: alloc_buffer=0,ndata_prev=0
#else
    real(dp),allocatable :: buffs(:),buffr(:)
#endif
    iup = itable(myranki+1,myrankj,myrankk)
    idown = itable(myranki-1,myrankj,myrankk)
    inum = iend-ista+1; jnum = jend-jsta+1; knum = kend-ksta+1
    ndata = jnum*knum
#ifdef ES
    if (alloc_buffer /= 0) then
       if (ndata > ndata_prev) then
          deallocate(buffs,buffr)
          allocate(buffs(ndata),buffr(ndata))
          ndata_prev = ndata
       end if
    else
       allocate(buffs(ndata),buffr(ndata))
       alloc_buffer = 1
       ndata_prev = ndata
    end if
#else
    allocate(buffs(ndata),buffr(ndata))
#endif
    if (myranki /= iprocs-1) then
       do k = 1,knum
          do j = 1,jnum
             buffs(j+(k-1)*jnum) = x(inum,j,k)
          end do
       end do
       call mpi_send(buffs,ndata,mpi_double_precision,iup,  1, &
            mpi_comm_world,ierr)
    endif
    if (myranki /= 0) then
       call mpi_recv(buffr,ndata,mpi_double_precision,idown,1, &
            mpi_comm_world,istatus,ierr)
       do k = 1,knum
          do j = 1,jnum
             x(0,j,k) = buffr(j+(k-1)*jnum)
          end do
       end do
    end if
#ifndef ES
    deallocate(buffs,buffr)
#endif
    return
  end subroutine mypartialshift1l1
! --------------------------------------------------------------------
  subroutine mypartialshift1l2(x,ista,iend,jsta,jend,ksta,kend)
!*********************************************************************
! send to idonw | recv from iup
!*********************************************************************
    integer :: i,j,k
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: x(0:iend-ista+1+1,0:jend-jsta+1+1,0:kend-ksta+1+1)
    integer :: inum,jnum,knum,iup,idown,ireqs,ireqr,ndata
#ifdef ES
    real(dp),allocatable,save :: buffs(:),buffr(:)
    integer,save :: alloc_buffer=0,ndata_prev=0
#else
    real(dp),allocatable :: buffs(:),buffr(:)
#endif
    iup = itable(myranki+1,myrankj,myrankk)
    idown = itable(myranki-1,myrankj,myrankk)
    inum = iend-ista+1; jnum = jend-jsta+1; knum = kend-ksta+1
    ndata = jnum*knum
#ifdef ES
    if (alloc_buffer /= 0) then
       if (ndata > ndata_prev) then
          deallocate(buffs,buffr)
          allocate(buffs(ndata),buffr(ndata))
          ndata_prev = ndata
       end if
    else
       allocate(buffs(ndata),buffr(ndata))
       alloc_buffer = 1
       ndata_prev = ndata
    end if
#else
    allocate(buffs(ndata),buffr(ndata))
#endif
    if (myranki /= 0) then
       do k = 1,knum
          do j = 1,jnum
             buffs(j+(k-1)*jnum) = x(1,j,k)
          end do
       end do
       call mpi_send(buffs,ndata,mpi_double_precision,idown,1, &
            mpi_comm_world,ierr)
    end if
    if (myranki /= iprocs-1) then
       call mpi_recv(buffr,ndata,mpi_double_precision,iup,  1, &
            mpi_comm_world,istatus,ierr)
       do k = 1,knum
          do j = 1,jnum
             x(inum+1,j,k) = buffr(j+(k-1)*jnum)
          end do
       end do
    end if
#ifndef ES
    deallocate(buffs,buffr)
#endif
    return
  end subroutine mypartialshift1l2
! --------------------------------------------------------------------
  subroutine mypartialshift2l1(x,ista,iend,jsta,jend,ksta,kend)
!*********************************************************************
! send to jup | recv from jdown
!*********************************************************************
    integer :: i,j,k
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: x(0:iend-ista+1+1,0:jend-jsta+1+1,0:kend-ksta+1+1)
    integer :: inum,jnum,knum,jup,jdown,ireqs,ireqr,ndata
#ifdef ES
    real(dp),allocatable,save :: buffs(:),buffr(:)
    integer,save :: alloc_buffer=0, ndata_prev=0
#else
    real(dp),allocatable :: buffs(:),buffr(:)
#endif
    jup = itable(myranki,myrankj+1,myrankk)
    jdown = itable(myranki,myrankj-1,myrankk)
    inum = iend-ista+1; jnum = jend-jsta+1; knum = kend-ksta+1
    ndata = inum*knum
#ifdef ES
    if (alloc_buffer /= 0) then
       if (ndata > ndata_prev) then
          deallocate (buffs,buffr)
          allocate (buffs(ndata),buffr(ndata))
          ndata_prev = ndata
       end if
    else
       allocate (buffs(ndata),buffr(ndata))
       alloc_buffer = 1
       ndata_prev = ndata
    end if
#else
    allocate (buffs(ndata),buffr(ndata))
#endif
    if (myrankj /= jprocs-1) then
       do k = 1,knum
          do i = 1,inum
             buffs(i+(k-1)*inum) = x(i,jnum,k)
          end do
       end do
       call mpi_send(buffs,ndata,mpi_double_precision,jup,  1, &
            mpi_comm_world,ierr)
    endif
    if (myrankj /= 0) then
       call mpi_recv(buffr,ndata,mpi_double_precision,jdown,1, &
               mpi_comm_world,istatus,ierr)
       do k = 1,knum
          do i = 1,inum
             x(i,0,k) = buffr(i+(k-1)*inum)
          end do
       end do
    end if
#ifndef ES
    deallocate(buffs,buffr)
#endif
    return
  end subroutine mypartialshift2l1
! --------------------------------------------------------------------
  subroutine mypartialshift2l2(x,ista,iend,jsta,jend,ksta,kend)
!*********************************************************************
! send to jdown | recv from jup
!*********************************************************************
    integer :: i,j,k
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: x(0:iend-ista+1+1,0:jend-jsta+1+1,0:kend-ksta+1+1)
    integer :: inum,jnum,knum,jup,jdown,ireqs,ireqr,ndata
#ifdef ES
    real(dp),allocatable,save :: buffs(:),buffr(:)
    integer,save :: alloc_buffer=0, ndata_prev=0
#else
    real(dp),allocatable :: buffs(:),buffr(:)
#endif
    jup = itable(myranki,myrankj+1,myrankk)
    jdown = itable(myranki,myrankj-1,myrankk)
    inum = iend-ista+1; jnum = jend-jsta+1; knum = kend-ksta+1
    ndata = inum*knum
#ifdef ES
    if (alloc_buffer /= 0) then
       if (ndata > ndata_prev) then
          deallocate (buffs,buffr)
          allocate (buffs(ndata),buffr(ndata))
          ndata_prev = ndata
       end if
    else
       allocate (buffs(ndata),buffr(ndata))
       alloc_buffer = 1
       ndata_prev = ndata
    end if
#else
    allocate (buffs(ndata),buffr(ndata))
#endif
    if (myrankj /= 0) then
       do k = 1,knum
          do i = 1,inum
             buffs(i+(k-1)*inum) = x(i,1,k)
          end do
       end do
       call mpi_send(buffs,ndata,mpi_double_precision,jdown,1, &
            mpi_comm_world,ierr)
    endif
    if (myrankj /= jprocs-1) then
       call mpi_recv(buffr,ndata,mpi_double_precision,jup,  1, &
            mpi_comm_world,istatus,ierr)
       do k = 1,knum
          do i = 1,inum
             x(i,jnum+1,k) = buffr(i+(k-1)*inum)
          end do
       end do
    end if
#ifndef ES
    deallocate(buffs,buffr)
#endif
    return
  end subroutine mypartialshift2l2
! --------------------------------------------------------------------
  subroutine mypartialshift3l1(x,ista,iend,jsta,jend,ksta,kend)
!*********************************************************************
! send to kup | recv from kdown
!*********************************************************************
    integer :: i,j,k
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: x(0:iend-ista+1+1,0:jend-jsta+1+1,0:kend-ksta+1+1)
    integer :: inum,jnum,knum,kup,kdown,ireqs,ireqr,ndata
#ifdef ES
    real(dp),allocatable,save :: buffs(:),buffr(:)
    integer,save :: alloc_buffer=0,ndata_prev=0
#else
    real(dp),allocatable :: buffs(:),buffr(:)
#endif
    kup = itable(myranki,myrankj,myrankk+1)
    kdown = itable(myranki,myrankj,myrankk-1)
    inum = iend-ista+1; jnum = jend-jsta+1; knum = kend-ksta+1
    ndata = inum*jnum
#ifdef ES
    if (alloc_buffer /= 0) then
       if (ndata > ndata_prev) then
          deallocate(buffs,buffr)
          allocate(buffs(ndata),buffr(ndata))
          ndata_prev = ndata
       end if
    else
       allocate(buffs(ndata),buffr(ndata))
       alloc_buffer = 1
       ndata_prev = ndata
    end if
#else
    allocate(buffs(ndata),buffr(ndata))
#endif
    if (myrankk /= kprocs-1) then
       do j = 1,jnum
          do i = 1,inum
             buffs(i+(j-1)*inum) = x(i,j,knum)
          end do
       end do
       call mpi_send(buffs,ndata,mpi_double_precision,kup,  1, &
            mpi_comm_world,ierr)
    endif
    if (myrankk /= 0) then
       call mpi_recv(buffr,ndata,mpi_double_precision,kdown, &
            1,mpi_comm_world,istatus,ierr)
       do j = 1,jnum
          do i = 1,inum
             x(i,j,0) = buffr(i+(j-1)*inum)
          end do
       end do
    end if
#ifndef ES
    deallocate(buffs,buffr)
#endif
    return
  end subroutine mypartialshift3l1
! --------------------------------------------------------------------
  subroutine mypartialshift3l2(x,ista,iend,jsta,jend,ksta,kend)
!*********************************************************************
! send to kdown | recv from kup
!*********************************************************************
    integer :: i,j,k
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: x(0:iend-ista+1+1,0:jend-jsta+1+1,0:kend-ksta+1+1)
    integer :: inum,jnum,knum,kup,kdown,ireqs,ireqr,ndata
#ifdef ES
    real(dp),allocatable,save :: buffs(:),buffr(:)
    integer,save :: alloc_buffer=0,ndata_prev=0
#else
    real(dp),allocatable :: buffs(:),buffr(:)
#endif
    kup = itable(myranki,myrankj,myrankk+1)
    kdown = itable(myranki,myrankj,myrankk-1)
    inum = iend-ista+1; jnum = jend-jsta+1; knum = kend-ksta+1
    ndata = inum*jnum
#ifdef ES
    if (alloc_buffer /= 0) then
       if (ndata > ndata_prev) then
          deallocate(buffs,buffr)
          allocate(buffs(ndata),buffr(ndata))
          ndata_prev = ndata
       end if
    else
       allocate(buffs(ndata),buffr(ndata))
       alloc_buffer = 1
       ndata_prev = ndata
    end if
#else
    allocate(buffs(ndata),buffr(ndata))
#endif
    if (myrankk /= 0) then
       do j = 1,jnum
          do i = 1,inum
             buffs(i+(j-1)*inum) = x(i,j,1)
          end do
       end do
       call mpi_send(buffs,ndata,mpi_double_precision,kdown,1, &
            mpi_comm_world,ierr)
    endif
    if (myrankk /= kprocs-1) then
       call mpi_recv(buffr,ndata,mpi_double_precision,kup,  1, &
            mpi_comm_world,istatus,ierr)
       do j = 1,jnum
          do i = 1,inum
             x(i,j,knum+1) = buffr(i+(j-1)*inum)
          end do
       end do
    end if
#ifndef ES
    deallocate(buffs,buffr)
#endif
    return
  end subroutine mypartialshift3l2
! --------------------------------------------------------------------
  subroutine myshift123l(x,ista,iend,jsta,jend,ksta,kend)
!*********************************************************************
!*********************************************************************
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: x(0:iend-ista+1+1,0:jend-jsta+1+1,0:kend-ksta+1+1)
    call mypartialshift1l1(x,ista,iend,jsta,jend,ksta,kend)
    call mypartialshift1l2(x,ista,iend,jsta,jend,ksta,kend)
    call mypartialshift2l1(x,ista,iend,jsta,jend,ksta,kend)
    call mypartialshift2l2(x,ista,iend,jsta,jend,ksta,kend)
    call mypartialshift3l1(x,ista,iend,jsta,jend,ksta,kend)
    call mypartialshift3l2(x,ista,iend,jsta,jend,ksta,kend)
    return
  end subroutine myshift123l
! ---------------------------------------------------------------------
  subroutine myroundshift123l(x,ista,iend,jsta,jend,ksta,kend &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!*********************************************************************
    integer :: i,j,k,ij
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp) :: x(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1)
    integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
    integer :: iup,idown,jup,jdown,kup,kdown
    integer :: ireqs1,ireqs2,ireqr1,ireqr2
    integer :: ndatai,ndataj,ndatak,inum,jnum,knum
#ifdef ES
    real(dp),allocatable,save :: &
         bufsi1(:),bufsi2(:),bufri1(:),bufri2(:) &
         ,bufsj1(:),bufsj2(:),bufrj1(:),bufrj2(:) &
         ,bufsk1(:),bufsk2(:),bufrk1(:),bufrk2(:)
    integer,save :: alloc_buffer=0,ndatai_prev=0,ndataj_prev=0,ndatak_prev=0
#else
    real(dp),allocatable :: &
         bufsi1(:),bufsi2(:),bufri1(:),bufri2(:) &
         ,bufsj1(:),bufsj2(:),bufrj1(:),bufrj2(:) &
         ,bufsk1(:),bufsk2(:),bufrk1(:),bufrk2(:)
#endif
    iup = itable(myranki+1,myrankj,myrankk)
    idown = itable(myranki-1,myrankj,myrankk)
    jup = itable(myranki,myrankj+1,myrankk)
    jdown = itable(myranki,myrankj-1,myrankk)
    kup = itable(myranki,myrankj,myrankk+1)
    kdown = itable(myranki,myrankj,myrankk-1)
    inum = iend1-ista2+3; jnum = jend1-jsta2+3; knum = kend1-ksta2+3
    ndatai = jnum*knum; ndataj = knum*inum; ndatak = inum*jnum
    ! allocate buffer
#ifdef ES
    if (alloc_buffer /= 0) then
       if (ndatai > ndatai_prev) then
          deallocate(bufsi1,bufsi2,bufri1,bufri2)
          allocate(bufsi1(ndatai),bufsi2(ndatai), &
               bufri1(ndatai),bufri2(ndatai))
          ndatai_prev = ndatai
       end if
       if (ndataj > ndataj_prev) then
          deallocate(bufsj1,bufsj2,bufrj1,bufrj2)
          allocate(bufsj1(ndataj),bufsj2(ndataj), &
               bufrj1(ndataj),bufrj2(ndataj))
          ndataj_prev = ndataj
       end if
       if (ndatak > ndatak_prev) then
          deallocate(bufsk1,bufsk2,bufrk1,bufrk2)
          allocate(bufsk1(ndatak),bufsk2(ndatak), &
               bufrk1(ndatak),bufrk2(ndatak))
          ndatak_prev = ndatak
       end if
    else
       allocate(bufsi1(ndatai),bufsi2(ndatai),bufri1(ndatai),bufri2(ndatai) &
            ,bufsj1(ndataj),bufsj2(ndataj),bufrj1(ndataj),bufrj2(ndataj) &
            ,bufsk1(ndatak),bufsk2(ndatak),bufrk1(ndatak),bufrk2(ndatak))
       alloc_buffer = 1
       ndatai_prev = ndatai
       ndataj_prev = ndataj
       ndatak_prev = ndatak
    end if
#else
    allocate(bufsi1(ndatai),bufsi2(ndatai),bufri1(ndatai),bufri2(ndatai) &
         ,bufsj1(ndataj),bufsj2(ndataj),bufrj1(ndataj),bufrj2(ndataj) &
         ,bufsk1(ndatak),bufsk2(ndatak),bufrk1(ndatak),bufrk2(ndatak))
#endif
    ! 1. shift in y-direction
    if (myrankj /= 0) then
       do k = ksta2-1,kend1+1
          do i = ista2-1,iend1+1
             ij = (i-ista2+2)+(k-ksta2+1)*inum
             bufsj1(ij) = x(i,jsta2,k)
          end do
       end do
       call mpi_send(bufsj1,ndataj,mpi_double_precision,jdown,1, &
            mpi_comm_world,ierr)
    end if
    if (myrankj /= jprocs-1) then
       call mpi_recv(bufrj1,ndataj,mpi_double_precision,jup,  1, &
            mpi_comm_world,istatus,ierr)
       do k = ksta2-1,kend1+1
          do i = ista2-1,iend1+1
             ij = (i-ista2+2)+(k-ksta2+1)*inum
             x(i,jend1+1,k) = bufrj1(ij)
          end do
       end do
    end if
    if (myrankj /= jprocs-1) then
       do k = ksta2-1,kend1+1
          do i = ista2-1,iend1+1
             ij = (i-ista2+2)+(k-ksta2+1)*inum
             bufsj2(ij) = x(i,jend1,k)
          end do
       end do
       call mpi_send(bufsj2,ndataj,mpi_double_precision,jup,  1, &
            mpi_comm_world,ierr)
    end if
    if (myrankj /= 0) then
       call mpi_recv(bufrj2,ndataj,mpi_double_precision,jdown,1, &
            mpi_comm_world,istatus,ierr)
       do k = ksta2-1,kend1+1
          do i = ista2-1,iend1+1
             ij = (i-ista2+2)+(k-ksta2+1)*inum
             x(i,jsta2-1,k) = bufrj2(ij)
          end do
       end do
    end if
    ! 2. shift in x-direction
    if (myranki /= 0) then
       do k = ksta2-1,kend1+1
          do j = jsta2-1,jend1+1
             ij = (j-jsta2+2)+(k-ksta2+1)*jnum
             bufsi1(ij) = x(ista2,j,k)
          end do
       end do
       call mpi_send(bufsi1,ndatai,mpi_double_precision,idown,1, &
            mpi_comm_world,ierr)
    end if
    if (myranki /= iprocs-1) then
       call mpi_recv(bufri1,ndatai,mpi_double_precision,iup,  1, &
            mpi_comm_world,istatus,ierr)
       do k = ksta2-1,kend1+1
          do j = jsta2-1,jend1+1
             ij = (j-jsta2+2)+(k-ksta2+1)*jnum
             x(iend1+1,j,k) = bufri1(ij)
          end do
       end do
    end if
    if (myranki /= iprocs-1) then
       do k = ksta2-1,kend1+1
          do j = jsta2-1,jend1+1
             ij = (j-jsta2+2)+(k-ksta2+1)*jnum
             bufsi2(ij) = x(iend1,j,k)
          end do
       end do
       call mpi_send(bufsi2,ndatai,mpi_double_precision,iup,  1, &
            mpi_comm_world,ierr)
    end if
    if (myranki /= 0) then
       call mpi_recv(bufri2,ndatai,mpi_double_precision,idown,1, &
            mpi_comm_world,istatus,ierr)
       do k = ksta2-1,kend1+1
          do j = jsta2-1,jend1+1
             ij = (j-jsta2+2)+(k-ksta2+1)*jnum
             x(ista2-1,j,k) = bufri2(ij)
          end do
       end do
    end if
    ! 3. shift in z-direction
    if (myrankk /= 0) then
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             ij = (i-ista2+2)+(j-jsta2+1)*inum
             bufsk1(ij) = x(i,j,ksta2)
          end do
       end do
       call mpi_send(bufsk1,ndatak,mpi_double_precision,kdown,1, &
            mpi_comm_world,ierr)
    end if
    if (myrankk /= kprocs-1) then
       call mpi_recv(bufrk1,ndatak,mpi_double_precision,kup,  1, &
            mpi_comm_world,istatus,ierr)
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             ij = (i-ista2+2)+(j-jsta2+1)*inum
             x(i,j,kend1+1) = bufrk1(ij)
          end do
       end do
    end if
    if (myrankk /= kprocs-1) then
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             ij = (i-ista2+2)+(j-jsta2+1)*inum
             bufsk2(ij) = x(i,j,kend1)
          end do
       end do
       call mpi_send(bufsk2,ndatak,mpi_double_precision,kup,  1, &
            mpi_comm_world,ierr)
    end if
    if (myrankk /= 0) then
       call mpi_recv(bufrk2,ndatak,mpi_double_precision,kdown,1, &
            mpi_comm_world,istatus,ierr)
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             ij = (i-ista2+2)+(j-jsta2+1)*inum
             x(i,j,ksta2-1) = bufrk2(ij)
          end do
       end do
    end if
#ifndef ES
    deallocate &
         (bufsi1,bufsi2,bufri1,bufri2 &
         ,bufsj1,bufsj2,bufrj1,bufrj2 &
         ,bufsk1,bufsk2,bufrk1,bufrk2)
#endif
    return
  end subroutine myroundshift123l
! ---------------------------------------------------------------------
  subroutine myroundshift123l4(x1,x2,x3,x4,ista,iend,jsta,jend,ksta,kend &
       ,ista2,iend1,jsta2,jend1,ksta2,kend1)
!*********************************************************************
!*********************************************************************
    integer :: i,j,k,ij
    integer,intent(in) :: ista,iend,jsta,jend,ksta,kend
    real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1),&
         intent(inout) :: x1,x2,x3,x4
    integer,intent(in) :: ista2,iend1,jsta2,jend1,ksta2,kend1
    integer :: iup,idown,jup,jdown,kup,kdown
    integer :: ireqs1,ireqs2,ireqr1,ireqr2
    integer :: isize,jsize,ksize,isize4,jsize4,ksize4
    integer :: inum,jnum,knum
#ifdef ES
    real(dp),allocatable,save :: &
         bufsi1(:),bufsi2(:),bufri1(:),bufri2(:), &
         bufsj1(:),bufsj2(:),bufrj1(:),bufrj2(:), &
         bufsk1(:),bufsk2(:),bufrk1(:),bufrk2(:)
    integer,save :: isize4_prev=0,jsize4_prev=0,ksize4_prev=0,alloc_buffer=0
#else
    real(dp),allocatable :: &
         bufsi1(:),bufsi2(:),bufri1(:),bufri2(:), &
         bufsj1(:),bufsj2(:),bufrj1(:),bufrj2(:), &
         bufsk1(:),bufsk2(:),bufrk1(:),bufrk2(:)
#endif
    iup = itable(myranki+1,myrankj,myrankk)
    idown = itable(myranki-1,myrankj,myrankk)
    jup = itable(myranki,myrankj+1,myrankk)
    jdown = itable(myranki,myrankj-1,myrankk)
    kup = itable(myranki,myrankj,myrankk+1)
    kdown = itable(myranki,myrankj,myrankk-1)
    inum = iend1-ista2+3; jnum = jend1-jsta2+3; knum = kend1-ksta2+3
    isize = jnum*knum; jsize = knum*inum; ksize = inum*jnum
    isize4=4*isize; jsize4=4*jsize; ksize4=4*ksize
    ! allocate communication buffers
#ifdef ES
    if (alloc_buffer /= 0) then ! already allocated
       if (isize4 > isize4_prev) then ! larger buffer than previous call
          deallocate(bufsi1,bufsi2,bufri1,bufri2)
          allocate(bufsi1(isize4),bufsi2(isize4),bufri1(isize4),bufri2(isize4))
#ifdef VERBOSE
          write(*,'("x",3i10)')myrank,isize4,isize4_prev
#endif
          isize4_prev=isize4
       endif
       if (jsize4 > jsize4_prev) then ! larger buffer than previous call
          deallocate(bufsj1,bufsj2,bufrj1,bufrj2)
          allocate(bufsj1(jsize4),bufsj2(jsize4),bufrj1(jsize4),bufrj2(jsize4))
#ifdef VERBOSE
          write(*,'("y",3i10)')myrank,jsize4,jsize4_prev
#endif
          jsize4_prev=jsize4
       endif
       if (ksize4 > ksize4_prev) then ! larger buffer than previous call
          deallocate(bufsk1,bufsk2,bufrk1,bufrk2)
          allocate(bufsk1(ksize4),bufsk2(ksize4),bufrk1(ksize4),bufrk2(ksize4))
#ifdef VERBOSE
          write(*,'("z",3i10)')myrank,ksize4,ksize4_prev
#endif
          ksize4_prev=ksize4
       endif
    else ! not yet allocated
       allocate &
            (bufsi1(isize4),bufsi2(isize4),bufri1(isize4),bufri2(isize4) &
            ,bufsj1(jsize4),bufsj2(jsize4),bufrj1(jsize4),bufrj2(jsize4) &
            ,bufsk1(ksize4),bufsk2(ksize4),bufrk1(ksize4),bufrk2(ksize4))
#ifdef VERBOSE
       write(*,'("xyz",7i10)')myrank,isize4,isize4_prev &
            ,jsize4,jsize4_prev,ksize4,ksize4_prev
#endif
       alloc_buffer = 1
       isize4_prev=isize4; jsize4_prev=jsize4; ksize4_prev=ksize4
    end if
#else
    allocate &
         (bufsi1(isize4),bufsi2(isize4),bufri1(isize4),bufri2(isize4) &
         ,bufsj1(jsize4),bufsj2(jsize4),bufrj1(jsize4),bufrj2(jsize4) &
         ,bufsk1(ksize4),bufsk2(ksize4),bufrk1(ksize4),bufrk2(ksize4))
#endif

    ! 1. shift in y-direction
    if (myrankj /= 0) then
       do k = ksta2-1,kend1+1
          do i = ista2-1,iend1+1
             ij = 4*((i-ista2+1)+(k-ksta2+1)*inum)
             bufsj1(ij+1) = x1(i,jsta2,k)
             bufsj1(ij+2) = x2(i,jsta2,k)
             bufsj1(ij+3) = x3(i,jsta2,k)
             bufsj1(ij+4) = x4(i,jsta2,k)
          end do
       end do
       call mpi_send(bufsj1,jsize4,mpi_double_precision,jdown,1, &
            mpi_comm_world,ierr)
    end if
    if (myrankj /= jprocs-1) then
       call mpi_recv(bufrj1,jsize4,mpi_double_precision,jup,  1, &
            mpi_comm_world,istatus,ierr)
       do k = ksta2-1,kend1+1
          do i = ista2-1,iend1+1
             ij = 4*((i-ista2+1)+(k-ksta2+1)*inum)
             x1(i,jend1+1,k) = bufrj1(ij+1)
             x2(i,jend1+1,k) = bufrj1(ij+2)
             x3(i,jend1+1,k) = bufrj1(ij+3)
             x4(i,jend1+1,k) = bufrj1(ij+4)
          end do
       end do
    end if
    if (myrankj /= jprocs-1) then
       do k = ksta2-1,kend1+1
          do i = ista2-1,iend1+1
             ij = 4*((i-ista2+1)+(k-ksta2+1)*inum)
             bufsj2(ij+1) = x1(i,jend1,k)
             bufsj2(ij+2) = x2(i,jend1,k)
             bufsj2(ij+3) = x3(i,jend1,k)
             bufsj2(ij+4) = x4(i,jend1,k)
          end do
       end do
       call mpi_send(bufsj2,jsize4,mpi_double_precision,jup,  1, &
            mpi_comm_world,ierr)
    end if
    if (myrankj /= 0) then
       call mpi_recv(bufrj2,jsize4,mpi_double_precision,jdown,1, &
            mpi_comm_world,istatus,ierr)
       do k = ksta2-1,kend1+1
          do i = ista2-1,iend1+1
             ij = 4*((i-ista2+1)+(k-ksta2+1)*inum)
             x1(i,jsta2-1,k) = bufrj2(ij+1)
             x2(i,jsta2-1,k) = bufrj2(ij+2)
             x3(i,jsta2-1,k) = bufrj2(ij+3)
             x4(i,jsta2-1,k) = bufrj2(ij+4)
          end do
       end do
    end if
    ! 2. shift in x-direction
    if (myranki /= 0) then
       do k = ksta2-1,kend1+1
          do j = jsta2-1,jend1+1
             ij = 4*((j-jsta2+1)+(k-ksta2+1)*jnum)
             bufsi1(ij+1) = x1(ista2,j,k)
             bufsi1(ij+2) = x2(ista2,j,k)
             bufsi1(ij+3) = x3(ista2,j,k)
             bufsi1(ij+4) = x4(ista2,j,k)
          end do
       end do
       call mpi_send(bufsi1,isize4,mpi_double_precision,idown,1, &
            mpi_comm_world,ierr)
    end if
    if (myranki /= iprocs-1) then
       call mpi_recv(bufri1,isize4,mpi_double_precision,iup,  1, &
            mpi_comm_world,istatus,ierr)
       do k = ksta2-1,kend1+1
          do j = jsta2-1,jend1+1
             ij = 4*((j-jsta2+1)+(k-ksta2+1)*jnum)
             x1(iend1+1,j,k) = bufri1(ij+1)
             x2(iend1+1,j,k) = bufri1(ij+2)
             x3(iend1+1,j,k) = bufri1(ij+3)
             x4(iend1+1,j,k) = bufri1(ij+4)
          end do
       end do
    end if
    if (myranki /= iprocs-1) then
       do k = ksta2-1,kend1+1
          do j = jsta2-1,jend1+1
             ij = 4*((j-jsta2+1)+(k-ksta2+1)*jnum)
             bufsi2(ij+1) = x1(iend1,j,k)
             bufsi2(ij+2) = x2(iend1,j,k)
             bufsi2(ij+3) = x3(iend1,j,k)
             bufsi2(ij+4) = x4(iend1,j,k)
          end do
       end do
       call mpi_send(bufsi2,isize4,mpi_double_precision,iup,  1, &
            mpi_comm_world,ierr)
    end if
    if (myranki /= 0) then
       call mpi_recv(bufri2,isize4,mpi_double_precision,idown,1, &
            mpi_comm_world,istatus,ierr)
       do k = ksta2-1,kend1+1
          do j = jsta2-1,jend1+1
             ij = 4*((j-jsta2+1)+(k-ksta2+1)*jnum)
             x1(ista2-1,j,k) = bufri2(ij+1)
             x2(ista2-1,j,k) = bufri2(ij+2)
             x3(ista2-1,j,k) = bufri2(ij+3)
             x4(ista2-1,j,k) = bufri2(ij+4)
          end do
       end do
    end if
    ! 3. shift in z-direction
    if (myrankk /= 0) then
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             ij = 4*((i-ista2+1)+(j-jsta2+1)*inum)
             bufsk1(ij+1) = x1(i,j,ksta2)
             bufsk1(ij+2) = x2(i,j,ksta2)
             bufsk1(ij+3) = x3(i,j,ksta2)
             bufsk1(ij+4) = x4(i,j,ksta2)
          end do
       end do
       call mpi_send(bufsk1,ksize4,mpi_double_precision,kdown,1, &
            mpi_comm_world,ierr)
    end if
    if (myrankk /= kprocs-1) then
       call mpi_recv(bufrk1,ksize4,mpi_double_precision,kup,  1, &
            mpi_comm_world,istatus,ierr)
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             ij = 4*((i-ista2+1)+(j-jsta2+1)*inum)
             x1(i,j,kend1+1) = bufrk1(ij+1)
             x2(i,j,kend1+1) = bufrk1(ij+2)
             x3(i,j,kend1+1) = bufrk1(ij+3)
             x4(i,j,kend1+1) = bufrk1(ij+4)
          end do
       end do
    end if
    if (myrankk /= kprocs-1) then
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             ij = 4*((i-ista2+1)+(j-jsta2+1)*inum)
             bufsk2(ij+1) = x1(i,j,kend1)
             bufsk2(ij+2) = x2(i,j,kend1)
             bufsk2(ij+3) = x3(i,j,kend1)
             bufsk2(ij+4) = x4(i,j,kend1)
          end do
       end do
       call mpi_send(bufsk2,ksize4,mpi_double_precision,kup,  1, &
            mpi_comm_world,ierr)
    end if
    if (myrankk /= 0) then
       call mpi_recv(bufrk2,ksize4,mpi_double_precision,kdown,1, &
            mpi_comm_world,istatus,ierr)
       do j = jsta2-1,jend1+1
          do i = ista2-1,iend1+1
             ij = 4*((i-ista2+1)+(j-jsta2+1)*inum)
             x1(i,j,ksta2-1) = bufrk2(ij+1)
             x2(i,j,ksta2-1) = bufrk2(ij+2)
             x3(i,j,ksta2-1) = bufrk2(ij+3)
             x4(i,j,ksta2-1) = bufrk2(ij+4)
          end do
       end do
    end if
#ifndef ES
    deallocate &
         (bufsi1,bufsi2,bufri1,bufri2 &
         ,bufsj1,bufsj2,bufrj1,bufrj2 &
         ,bufsk1,bufsk2,bufrk1,bufrk2)
#endif
    return
  end subroutine myroundshift123l4
! ---------------------------------------------------------------------
  function distributed_file_name(headname,irank)
    character(80),intent(in) :: headname
    integer,intent(in) :: irank
    character(87) :: distributed_file_name
    character(7) :: tailname
    integer :: i,length
    do i = 1,80
       if (headname(i:i).eq." ") exit
    end do
    length = i-1
    write(tailname,'(a,i4.4)')".p_",irank
    distributed_file_name = headname(1:length)//tailname
  end function distributed_file_name
! ---------------------------------------------------------------------
end module mympitools
