subroutine plot(tmpl,tsurf,tbot,qbot,ul,vl,wl,eivl,sivl,time,ntime)
  use constants; use mympitools; use mesh; use division
!**********************************************************************
!     plot average values
!**********************************************************************
  implicit none

  integer :: i,j,k,nn,ij

  real(dp),dimension(ista:iend,jsta:jend,ksta:kend) :: tmpl,ul,vl,wl,sivl,eivl
  real(dp) :: tsurf,tbot,time,qbot
  real(dp) :: vrms,vplt,tav,hfs,hfb,vrmsp,vpltp,tavp,hfsp,hfbp
  real(dp) :: smax,emax,smaxp,emaxp
  real(dp) :: dum,surfci,volumi
  integer :: ntime
  real(dp),dimension(ista-1:iend+1,jsta-1:jend+1,ksta-1:kend+1) :: uwk,vwk,wwk
  real(dp) :: as(7),ar(7*nprocs)
  surfci = 1.e0_dp/aspx/aspy
  volumi = surfci

! ================================
!     calculate avarage values
! ================================
!
! +++++++++++++++++++++++++++
!     average temperature
! +++++++++++++++++++++++++++

  tavp = 0.e0_dp
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           tavp = tavp + tmpl(i,j,k)*dx(i)*dy(j)*dz(k)
        enddo
     enddo
  enddo
  as(1) = tavp

! ++++++++++++++++++++
!     RMS velocity
! ++++++++++++++++++++

  uwk(ista:iend,jsta:jend,ksta:kend) = ul(ista:iend,jsta:jend,ksta:kend)
  vwk(ista:iend,jsta:jend,ksta:kend) = vl(ista:iend,jsta:jend,ksta:kend)
  wwk(ista:iend,jsta:jend,ksta:kend) = wl(ista:iend,jsta:jend,ksta:kend)
  call mypartialshift1l2(uwk,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift2l2(vwk,ista,iend,jsta,jend,ksta,kend)
  call mypartialshift3l2(wwk,ista,iend,jsta,jend,ksta,kend)

  vrmsp = 0.e0_dp
  do k = ksta2,kend1
     do j = jsta2,jend1
        do i = ista2,iend1
           vrmsp = vrmsp + dx(i)*dy(j)*dz(k)* &
                ((uwk(i,j,k)+uwk(i+1,j,k))**2 &
                +(vwk(i,j,k)+vwk(i,j+1,k))**2 &
                +(wwk(i,j,k)+wwk(i,j,k+1))**2)
        enddo
     enddo
  enddo
  vrmsp = 0.25e0_dp*vrmsp
  as(2) = vrmsp

! +++++++++++++++++++++++++++++++++++
!     RMS velocity on top surface
! +++++++++++++++++++++++++++++++++++

  vpltp = 0.e0_dp
  if (myrankk == kprocs-1) then
     k = kax
     do j = jsta2,jend1
        do i = ista2,iend1
           dum = (uwk(i,j,k-1)+uwk(i+1,j,k-1) &
                +uwk(i,j,k)+uwk(i+1,j,k))**2 &
                +(vwk(i,j,k-1)+vwk(i,j+1,k-1) &
                +vwk(i,j,k)+vwk(i,j+1,k))**2
           vpltp = vpltp + dum*dx(i)*dy(j)
        enddo
     enddo
  endif
  vpltp = .0625e0_dp*vpltp
  as(3) = vpltp

! ++++++++++++++++++++++++++++++++
!     heat flow at top surface
! ++++++++++++++++++++++++++++++++

  hfsp = 0.e0_dp
  if (myrankk == kprocs-1) then
     do j = jsta2,jend1
        do i = ista2,iend1
           hfsp = hfsp + 2.e0_dp*(tmpl(i,j,kax-1)-tsurf)*dx(i)*dy(j)/dz(kax-1)
        enddo
     enddo
  endif
  as(4) = hfsp

! +++++++++++++++++++++++++++++++++++
!     heat flow at bottom surface
! +++++++++++++++++++++++++++++++++++

  hfbp = 0.e0_dp
  if (myrankk == 0) then
     do j = jsta2,jend1
        do i = ista2,iend1
           hfbp = hfbp + 2.e0_dp*(tbot-tmpl(i,j,2))*dx(i)*dy(j)/dz(2)
        enddo
     enddo
  endif
  as(5) = hfbp
!
! ================================
!     calculate maximum values
! ================================
!
  smaxp = maxval(sivl(ista2:iend1,jsta2:jend1,ksta2:kend1))
  emaxp = maxval(eivl(ista2:iend1,jsta2:jend1,ksta2:kend1))
  as(6) = smaxp
  as(7) = emaxp

  call mpi_allgather(as,7,mpi_double_precision, &
       ar,7,mpi_double_precision,mpi_comm_world,ierr)
  tav = sum(ar(1:7*nprocs:7))*volumi
  vrms = dsqrt(sum(ar(2:7*nprocs:7))*volumi)
  vplt = dsqrt(sum(ar(3:7*nprocs:7))*surfci)
  hfs = sum(ar(4:7*nprocs:7))*surfci
  hfb = sum(ar(5:7*nprocs:7))*surfci
  smax = maxval(ar(6:7*nprocs:7))
  emax = maxval(ar(7:7*nprocs:7))

  if (myrank == 0) then
     write(24,'(i6,a4,10a10)')ntime,"time","vrms","vplt","tav", &
          "hfs","hfb","tbot","smax","emax"
     write(24,'(3e10.4,4f10.6,4e10.4)')time,vrms,vplt,tav, &
          hfs,hfb,tbot,smax,emax
     write(20,'(11e11.4)')time,vrms,vplt,tav,hfs,hfb,tbot,smax,emax
  end if

  return
end subroutine plot
