subroutine rdata(nmax,nstep,iplot,iprnt,time,ivlt,ivlb,icont, &
     ivis,ipt,etl,etn,epl,epn,rat,tsurf,tbot,tamp,vboti, &
     ipf,relu,relv,relw,relp,crmax)
  use constants; use mesh; use mympitools
!**********************************************************************
!     read input parameters
!**********************************************************************
  implicit none

  integer :: nmax,iplot,iprnt,nstep
  integer :: ivlt,ivlb,icont
  real(dp) :: rat,tsurf,tbot,tamp,vboti
  real(dp) :: time
  integer :: ipt,ivis
  real(dp) :: etl,etn,epl,epn
  real(dp),parameter :: depth=6.5e5_dp, thex=3.e-5_dp, grav=9.8e0_dp, &
       thdif=6.e-7_dp, rhoref=3.3e3_dp, rgas=8.31e0_dp, tgap = 1.3e3_dp
  real(dp) :: eta0,etas,etab,strs0
  real(dp) :: dum,duml,dumn
  real(dp),parameter :: eps=1.e-20_dp
  character(80) :: ipf
  real(dp) :: relu,relv,relw,relp,crmax
  logical :: unsupported

  open(21,file=ipf)
  read(21,*) nmax,iplot,iprnt,icont
  read(21,*) rat,aspx,aspy,ivlt,ivlb
  read(21,*) ipt,etl,etn,epl,epn
  read(21,*) tsurf,tbot,tamp,vboti
  read(21,*) nstep,time
  read(21,*) relu,relv,relw,relp,crmax
  close(21)

!     show some parameters in real dimensions

  eta0 = rhoref*thex*tgap*grav*depth**3/thdif/rat
  strs0 = eta0*thdif/depth/depth*1.e-6_dp
  if (ipt == 0) then
     etas = 1.e0_dp
     etab = 1.e0_dp
  else if (ipt == 1) then
     etas = exp(etl/tsurf)
     etab = exp((etl+epl)/tbot)
  else if (ipt == 2) then
     etas = exp(-etl*tsurf)
     etab = exp(-etl*tbot+epl)
  endif

!     check whether options are supported or not
  unsupported = .false.
  ! Arhenius-law dependence of viscosity
  if (ipt == 1) unsupported = .true.
  ! illegal ipt
  if (ipt >= 3) unsupported = .true.
  ! nonlinear-dependence of viscosity
  duml = etl**2+epl**2; dumn = etn**2+epn**2
  dum = duml+dumn
  if (dum <= eps) then
     ivis = 0
  else if (dumn <=  eps) then
     ivis = 1
  else
     ivis = 2
     unsupported = .true.
  endif

  if (myrank == 0) then
     write(24,'(4a10)') "nmax","iplot","iprnt","icont"
     write(24,'(4i10)') nmax,iplot,iprnt,icont
     write(24,'(5a10)')"Nx","Ny","Nz","aspx","aspy"
     write(24,'(3i10,2f10.4)')nx,ny,nz,aspx,aspy
     write(24,'(4a10)') "RaT","ivlt","ivlb","ipt"
     write(24,'(e10.4,3i10)') rat,ivlt,ivlb,ipt
     write(24,'(7a10)')"eta0","etas","etab","etl","etn","epl","epn"
     write(24,'(7e10.4)')eta0,etas,etab,etl,etn,epl,epn
     write(24,'(4a10)') "tsurf","tbot","tamp","vbot^(-1)"
     write(24,'(4f10.4)') tsurf,tbot,tamp,vboti
     write(24,'(a10)') "time"
     write(24,'(e10.4)') time
     write(24,'(5a10)') "relu","relv","relw","relp","crmax"
     write(24,'(4f10.4,e10.4)') relu,relv,relw,relp,crmax
     if (ivlt == 1) write(24,*)"free-slip top boundary"
     if (ivlt == 2) write(24,*)"fixed velocity top boundary"
     if (ivlb == 1) write(24,*)"free-slip bottom boundary"
     if (ivlb == 2) write(24,*)"fixed velocity bottom boundary"
     if (ipt == 1) write(24,*)"DO NOT USE ipt=1; Arhenius-law dependence"
     if (ipt == 2) write(24,*)"simplified P/T dependence"
     if (ipt >= 3) write(24,'("Illegal value of ipt =",i2)')ipt
     if (ivis == 0) write(24,*)"constant viscosity"
     if (ivis == 1) write(24,*)"T/P-dependent linear rheology"
     if (ivis == 2) write(24,*)"Nonlinear rheology; currently not available"
  endif

  if (unsupported) stop
  return
end subroutine rdata
